# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch
import torchvision
from torch.utils.data import Sampler, TensorDataset

from renate.memory.buffer import ReservoirBuffer
from renate.utils import pytorch
from renate.utils.pytorch import (
    ConcatRandomSampler,
    cat_nested_tensors,
    complementary_indices,
    get_length_nested_tensors,
    randomly_split_data,
    unique_classes,
)

from dummy_datasets import DummyDataIncrementalDataModule


@pytest.mark.parametrize("model", [torchvision.models.resnet18(pretrained=True)])
def test_reinitialize_model_parameters(model: torch.nn.Module):
    model.train()
    model(torch.randn(10, 3, 128, 128))  # Peform a forward pass, to change batchnorm buffers.
    params_before = [p.clone() for p in model.parameters()]
    buffers_before = [b.clone() for b in model.buffers()]
    pytorch.reinitialize_model_parameters(model)
    for p_before, p_after in zip(params_before, model.parameters()):
        assert not torch.allclose(p_before, p_after)
    for b_before, b_after in zip(buffers_before, model.buffers()):
        assert not torch.allclose(b_before, b_after)


@pytest.mark.parametrize(
    "proportions,expected_sizes",
    [
        [[0.5, 0.5], [50, 50]],
        [[0.3, 0.3, 0.4], [30, 30, 40]],
        [[0.333, 0.333, 0.334], [33, 33, 34]],
    ],
)
@pytest.mark.parametrize("splitter_func", [randomly_split_data])
def test_split_size(proportions, expected_sizes, splitter_func):
    dataset = TensorDataset(torch.randn(100, 1))
    splits = splitter_func(dataset, proportions, len(dataset))
    for i, split in enumerate(splits):
        assert len(split) == expected_sizes[i]


@pytest.mark.parametrize("splitter_func", [randomly_split_data])
def test_split_not_overlapping_splits(splitter_func):
    X = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    dataset = TensorDataset(X)
    splits = splitter_func(dataset, [0.5, 0.25, 0.25], len(dataset))
    for i in range(len(splits)):
        for j in range(len(splits)):
            if i == j:
                continue
            for k in range(len(splits[i])):
                assert splits[i][k] not in splits[j]


def test_random_splitting_sample_split_with_same_random_seed():
    X = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    dataset = TensorDataset(X)

    d_1_split_1, d_1_split_2 = randomly_split_data(dataset, [0.5, 0.5], 42)
    d_2_split_1, d_2_split_2 = randomly_split_data(dataset, [0.5, 0.5], 42)

    for i in range(5):
        assert torch.equal(d_1_split_1[i][0], d_2_split_1[i][0])
        assert torch.equal(d_1_split_2[i][0], d_2_split_2[i][0])


def test_get_length_nested_tensors():
    expected_batch_size = 10
    t = torch.zeros(expected_batch_size)
    assert get_length_nested_tensors(t) == expected_batch_size
    tuple_tensor = (t, t)
    assert get_length_nested_tensors(tuple_tensor) == expected_batch_size
    dict_tensor = {"k1": t, "k2": t}
    assert get_length_nested_tensors(dict_tensor) == expected_batch_size


def test_cat_nested_tensors():
    tensor_dim = 2
    first_dim_ones = 8
    zeros = torch.zeros((2, tensor_dim))
    ones = torch.ones((first_dim_ones, tensor_dim))
    result = cat_nested_tensors((zeros, ones))
    assert get_length_nested_tensors(result) == 10
    assert result.mean() == 0.8
    tuple_tensor = (zeros, ones)
    result = cat_nested_tensors((tuple_tensor, tuple_tensor))
    assert get_length_nested_tensors(result) == 4
    assert result[0].sum() == 0
    assert result[1].sum() == 2 * first_dim_ones * tensor_dim
    dict_tensor = {"zeros": zeros, "ones": ones}
    result = cat_nested_tensors((dict_tensor, dict_tensor))
    assert get_length_nested_tensors(result) == 4
    assert result["zeros"].sum() == 0
    assert result["ones"].sum() == 2 * first_dim_ones * tensor_dim


def test_cat_nested_tensors_wrong_shape():
    tensor1 = torch.zeros((2, 2))
    tensor2 = torch.zeros((2, 3))
    with pytest.raises(RuntimeError, match=r"Sizes of tensors must match except in dimension 0.*"):
        cat_nested_tensors((tensor1, tensor2))
    with pytest.raises(RuntimeError, match=r"Sizes of tensors must match except in dimension 0.*"):
        cat_nested_tensors(((tensor1, tensor1), (tensor1, tensor2)))
    with pytest.raises(RuntimeError, match=r"Sizes of tensors must match except in dimension 0.*"):
        cat_nested_tensors(({"k1": tensor1, "k2": tensor1}, {"k1": tensor1, "k2": tensor2}))


@pytest.mark.parametrize(
    "num_outputs, indices, expected_output",
    [
        [5, {2, 4}, [0, 1, 3]],
        [torch.rand(5, 5).size(1), {1, 2, 3}, [0, 4]],
        [torch.rand(5, 5).shape[1], {1, 2, 3}, [0, 4]],
    ],
)
def test_complementary_indices(num_outputs, indices, expected_output):
    assert expected_output == complementary_indices(num_outputs, indices)


@pytest.mark.parametrize("test_dataset", [True, False])
def test_unique_classes(tmpdir, test_dataset):
    if test_dataset:
        for data_id in range(5):
            data_module = DummyDataIncrementalDataModule(
                data_id, (10, 10), transform=None, val_size=0
            )
            data_module.prepare_data()
            data_module.setup()
            train_data = data_module.train_data()
            predicted_unique = set(int(x) for x in unique_classes(train_data))
            assert predicted_unique == {data_id}
    else:
        X = torch.randn(10, 3)
        y = torch.arange(0, 10)
        ds = torch.utils.data.TensorDataset(X, y)
        metadata = {"foo": torch.ones(10)}
        buffer = ReservoirBuffer(X.shape[0])
        buffer.update(ds, metadata)
        predicted_unique = unique_classes(buffer)
        assert predicted_unique == set(list(range(10)))


@pytest.mark.parametrize(
    "complete_dataset_iteration,expected_batches", [[None, 2], [0, 7], [1, 5], [2, 2]]
)
def test_concat_random_sampler(complete_dataset_iteration, expected_batches):
    sampler = ConcatRandomSampler(
        dataset_lengths=[15, 5, 20],
        batch_sizes=[2, 1, 8],
        complete_dataset_iteration=complete_dataset_iteration,
    )
    assert len(sampler) == expected_batches
    num_batches = 0
    for sample in sampler:
        assert all([s < 15 for s in sample[:2]])
        assert all([15 <= s < 20 for s in sample[2:3]])
        assert all([20 <= s < 40 for s in sample[3:]])
        num_batches += 1
    assert num_batches == expected_batches


def test_concat_random_sampler_distributed():
    """Tests behavior in case of distributed computing."""
    mock_sampler = Sampler(None)
    mock_sampler.rank = 1
    mock_sampler.num_replicas = 2
    expected_batches = 2
    sampler = ConcatRandomSampler(
        dataset_lengths=[16, 10], batch_sizes=[2, 2], sampler=mock_sampler
    )
    assert len(sampler) == expected_batches
    num_batches = 0
    for sample in sampler:
        assert all([7 < s < 16 for s in sample[:2]])
        assert all([21 <= s < 26 for s in sample[2:]])
        num_batches += 1
    assert num_batches == expected_batches



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pickle

import pytest
import torch
from avalanche.training.plugins import EWCPlugin, ReplayPlugin
from torch import Tensor
from torch.utils.data import Subset, TensorDataset

from renate.memory import ReservoirBuffer
from renate.utils.avalanche import (
    AvalancheBenchmarkWrapper,
    _plugin_index,
    plugin_by_class,
    remove_plugin,
    replace_plugin,
    to_avalanche_dataset,
)


@pytest.mark.parametrize("use_buffer", [True, False])
def test_to_avalanche_dataset(use_buffer):
    expected_x = 6
    expected_y = 1
    tensor_dataset = TensorDataset(
        torch.tensor([5, expected_x, 7]), torch.tensor([0, expected_y, 2])
    )
    if use_buffer:
        buffer = ReservoirBuffer(10)
        buffer.update(tensor_dataset)
        dataset = to_avalanche_dataset(buffer)
    else:
        dataset = to_avalanche_dataset(Subset(tensor_dataset, [0, 1, 2]))
    assert type(dataset._targets) == list
    assert len(dataset._targets) == 3
    assert dataset._targets[1] == expected_y
    assert type(dataset._targets[0]) == int
    assert dataset.targets[0].item() == dataset._targets[0]
    assert dataset.targets[1] == expected_y
    assert type(dataset.targets) == Tensor
    x, y = dataset[1]
    assert x == expected_x and y == expected_y
    assert len(dataset) == 3
    pickle.dumps(dataset)


def test_avalanche_benchmark_wrapper_correctly_tracks_and_saves_state():
    """Check if the state we need to track for Avalanche works correctly."""

    def get_benchmark(dataset):
        return AvalancheBenchmarkWrapper(
            train_dataset=dataset,
            val_dataset=dataset,
            train_transform=None,
            train_target_transform=None,
            test_transform=None,
            test_target_transform=None,
        )

    dataset = TensorDataset(Tensor([5, 5]), Tensor([0, 1]))
    benchmark = get_benchmark(dataset)
    benchmark.update_benchmark_properties()
    assert benchmark._n_classes_per_exp == [2]
    assert benchmark._classes_order == [0, 1]
    benchmark_state = benchmark.state_dict()
    dataset = TensorDataset(Tensor([5, 5]), Tensor([3, 2]))
    benchmark = get_benchmark(dataset)
    benchmark.load_state_dict(benchmark_state)
    assert benchmark._n_classes_per_exp == [2]
    assert benchmark._classes_order == [0, 1]
    benchmark.update_benchmark_properties()
    assert benchmark._n_classes_per_exp == [2, 2]
    assert benchmark._classes_order == [0, 1, 2, 3]


def test_given_plugin_exists_in_list_replace_it():
    """Tests if `replace_plugin` replaces a plugin if a plugin of this class already exists."""
    to_be_replaced = ReplayPlugin()
    to_be_added = ReplayPlugin()
    plugins = [to_be_replaced, EWCPlugin(ewc_lambda=1)]
    plugins = replace_plugin(to_be_added, plugins)
    assert to_be_added == plugins[0]
    assert to_be_replaced not in plugins


def test_given_plugin_not_exists_in_list_append_it():
    """Tests if `replace_plugin` appends a plugin if a plugin of this class does not exist."""
    to_be_added = ReplayPlugin()
    plugins = [EWCPlugin(ewc_lambda=1)]
    plugins = replace_plugin(to_be_added, plugins)
    assert to_be_added == plugins[-1]


def test_given_type_and_object_exists_in_list_returns_it():
    """``plugin_by_class`` returns object by type from list if exists."""
    expected = ReplayPlugin()
    assert expected == plugin_by_class(ReplayPlugin, [expected, EWCPlugin(ewc_lambda=1)])


def test_given_plugin_exists_in_list_remove_it():
    """Tests if `remove_plugin` removes a plugin if a plugin of this class already exists."""
    to_be_removed = ReplayPlugin()
    expected_plugin = EWCPlugin(ewc_lambda=1)
    plugins = [to_be_removed, expected_plugin]
    plugins = remove_plugin(ReplayPlugin, plugins)
    assert expected_plugin == plugins[0]
    assert len(plugins) == 1


def test_given_type_and_object_exists_not_in_list_returns_none():
    """``plugin_by_class`` returns None if type is not in list."""
    assert None is plugin_by_class(ReplayPlugin, [EWCPlugin(ewc_lambda=1)])


def test_duplicate_plugins_raise_exception():
    """Helper function do not allow for duplicate plugin types."""
    with pytest.raises(ValueError):
        _plugin_index(ReplayPlugin, [ReplayPlugin(), ReplayPlugin()])



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import datetime
import os
from typing import Any, Dict, Literal

from pytorch_lightning.loggers import TensorBoardLogger
from syne_tune.optimizer.schedulers import FIFOScheduler

OPTIMIZER = "Adam"
SUPPORTED_OPTIMIZERS = ["Adam", "SGD"]
SUPPORTED_OPTIMIZERS_TYPE = Literal["Adam", "SGD"]
LR_SCHEDULER_INTERVAL = "epoch"
SUPPORTED_LR_SCHEDULER_INTERVAL = ["epoch", "step"]
SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = Literal["epoch", "step"]
LEARNING_RATE = 3e-4
MOMENTUM = 0.0
WEIGHT_DECAY = 0.0
MAX_EPOCHS = 50
BATCH_SIZE = 32
BATCH_MEMORY_FRAC = 0.5
LOSS_WEIGHT = 1.0
SEED = 0
EMA_MEMORY_UPDATE_GAMMA = 1.0
VALIDATION_SIZE = 0.0
LOSS_NORMALIZATION = 1
EARLY_STOPPING = False
DETERMINISTIC_TRAINER = False

ACCELERATOR = "auto"
SUPPORTED_ACCELERATORS = ["auto", "cpu", "gpu", "tpu"]
SUPPORTED_ACCELERATORS_TYPE = Literal["auto", "cpu", "gpu", "tpu"]
DEVICES = 1
VOLUME_SIZE = 60
DISTRIBUTED_STRATEGY = "ddp"
PRECISION = "32"
GRADIENT_CLIP_VAL = None
GRADIENT_CLIP_ALGORITHM = None

LEARNER = "ER"
INSTANCE_COUNT = 1
INSTANCE_MAX_TIME = 3 * 24 * 3600
N_WORKERS = 1
INSTANCE_TYPE = "ml.c5.xlarge"
PYTHON_VERSION = "py39"
FRAMEWORK_VERSION = "1.13.1"

TASK_ID = "default_task"
MASK_UNUSED_CLASSES = False
WORKING_DIRECTORY = "renate_working_dir"
LOGGER = TensorBoardLogger
LOGGER_KWARGS = {
    "save_dir": os.path.expanduser(os.path.join("~", ".renate", WORKING_DIRECTORY)),
    "version": 1,
    "name": "lightning_logs",
}
JOB_KWARGS_FILE = "job_kwargs.json"
JOB_NAME = "renate"
SUPPORTED_TUNING_MODE = ["min", "max"]
SUPPORTED_TUNING_MODE_TYPE = Literal["min", "max"]
SAVE_BENCHMARK_STATE = True

SUPPORTED_BACKEND = ["local", "sagemaker"]
SUPPORTED_BACKEND_TYPE = Literal["local", "sagemaker"]

# ER
ER_ALPHA = 1.0

# DER
DER_ALPHA = 1.0
DER_BETA = 1.0

# POD-ER
POD_ALPHA = 1.0
POD_DISTILLATION_TYPE = "spatial"
POD_NORMALIZE = 1

# CLS-ER
CLS_ALPHA = 0.5
CLS_BETA = 0.1
CLS_STABLE_MODEL_UPDATE_WEIGHT = 0.999
CLS_STABLE_MODEL_UPDATE_PROBABILITY = 0.7
CLS_PLASTIC_MODEL_UPDATE_WEIGHT = 0.999
CLS_PLASTIC_MODEL_UPDATE_PROBABILITY = 0.9

# SuperER
SER_DER_ALPHA = 1.0
SER_DER_BETA = 1.0
SER_SP_SHRINK_FACTOR = 0.95
SER_SP_SIGMA = 0.001
SER_POD_ALPHA = 1.0
SER_POD_DISTILLATION_TYPE = "spatial"
SER_POD_NORMALIZE = 1
SER_CLS_ALPHA = 0.1
SER_CLS_STABLE_MODEL_UPDATE_WEIGHT = 0.999
SER_CLS_STABLE_MODEL_UPDATE_PROBABILITY = 0.7
SER_CLS_PLASTIC_MODEL_UPDATE_WEIGHT = 0.999
SER_CLS_PLASTIC_MODEL_UPDATE_PROBABILITY = 0.9

# EWC
EWC_LAMBDA = 0.4

# LwF
LWF_ALPHA = 1
LWF_TEMPERATURE = 2

MEMORY_SIZE = 32

# Benchmark datasets/models
TOKENIZER_KWARGS = {"padding": "max_length", "max_length": 128, "truncation": True}

# L2p
PROMPT_SIM_LOSS_WEIGHT = 0.5

# S-prompt
CLUSTERS_PER_TASK = 5


def scheduler(config_space: Dict[str, Any], mode: str, metric: str):
    return FIFOScheduler(
        config_space=config_space,
        searcher="random",
        mode=mode,
        metric=metric,
    )


def current_timestamp() -> str:
    return str(datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S-%f"))


def data_folder(working_directory: str) -> str:
    return os.path.join(working_directory, "data")


def input_state_folder(working_directory: str) -> str:
    return os.path.join(working_directory, "input_state")


def output_state_folder(working_directory: str) -> str:
    return os.path.join(working_directory, "output_state")


def logs_folder(working_directory: str) -> str:
    return os.path.join(working_directory, "logs")


def model_file(state_folder: str) -> str:
    return os.path.join(state_folder, "model.ckpt")


LEARNER_CHECKPOINT_NAME = "learner.ckpt"
AVALANCHE_CHECKPOINT_NAME = "avalanche.ckpt"


def learner_state_file(state_folder: str) -> str:
    return os.path.join(state_folder, LEARNER_CHECKPOINT_NAME)


def avalanche_state_file(state_folder: str) -> str:
    return os.path.join(state_folder, AVALANCHE_CHECKPOINT_NAME)


def metric_summary_file(logs_folder: str, special_str: str = "") -> str:
    return os.path.join(logs_folder, f"metrics_summary{special_str}.csv")


def hpo_file(state_folder: str) -> str:
    return os.path.join(state_folder, "hpo.csv")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import math
import os
from pathlib import Path
from typing import Any, Optional, Tuple, Union
from warnings import warn

import torch

from renate.types import NestedTensors


def mmap_tensor(
    filename: str, size: Union[int, Tuple[int, ...]], dtype: torch.dtype
) -> torch.Tensor:
    """Creates or accesses a memory-mapped tensor."""
    t = torch.from_file(
        filename,
        shared=True,
        size=math.prod(size) if isinstance(size, tuple) else size,
        dtype=dtype,
        device="cpu",
    )
    return t.view(size)


class Storage(torch.utils.data.Dataset):
    """An abstract class for permanent storage of datasets."""

    def __init__(self, directory: str) -> None:
        self._directory = directory

    def __len__(self) -> int:
        return self._length

    def __getitem__(self, idx: int) -> Any:
        raise NotImplementedError()

    def dump_dataset(self, ds: torch.utils.data.Dataset) -> None:
        raise NotImplementedError()

    def load_dataset(self, directory: Union[str, Path]):
        raise NotImplementedError()


class MemoryMappedTensorStorage(Storage):
    """A class implementing permanent storage of nested tensor datasets.

    This implements storage for `length` data points consisting of nested tensors of fixed types
    and shapes. `Storage` implements `__len__` and `__getitem__` and therefore can be used as a
    torch `Dataset`. To populate the storage, it also implements `dump_dataset`. It does _not_ keep
    track which slots have or have not been populated.

    `Storage` is given a path to a directory, where it creates (or accesses, if they already exist)
    memory-mapped tensor files.

    Args:
        directory: Path to a directory.
        data_point: Prototypical datapoint from which to infer shapes/dtypes.
        length: Number of items to be stored.
    """

    def __init__(self, directory: str) -> None:
        warn(
            f"""{self.__class__.__name__} will be deprecated very soon. Use FileTensorStorage
            instead. {self.__class__.__name__} is currently not fully functional, as some of the
            necessary parts of the interface have been modified and simplified. """,
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(directory)
        self._storage: Optional[NestedTensors] = None

    @staticmethod
    def _create_mmap_tensors(path: str, data_point: NestedTensors, length: int) -> NestedTensors:
        if isinstance(data_point, torch.Tensor):
            os.makedirs(os.path.dirname(path), exist_ok=True)
            filename = f"{path}.pt"
            return mmap_tensor(filename, size=(length, *data_point.size()), dtype=data_point.dtype)
        elif isinstance(data_point, tuple):
            return tuple(
                MemoryMappedTensorStorage._create_mmap_tensors(
                    os.path.join(path, f"{i}.pt"), data_point[i], length
                )
                for i in range(len(data_point))
            )
        elif isinstance(data_point, dict):
            return {
                key: MemoryMappedTensorStorage._create_mmap_tensors(
                    os.path.join(path, f"{key}.pt"), data_point[key], length
                )
                for key in data_point
            }
        else:
            raise TypeError(f"Expected nested tuple/dict of tensors, found {type(data_point)}.")

    @staticmethod
    def _get(storage: NestedTensors, idx: int) -> NestedTensors:
        if isinstance(storage, torch.Tensor):
            return storage[idx]
        elif isinstance(storage, tuple):
            return tuple(MemoryMappedTensorStorage._get(t, idx) for t in storage)
        elif isinstance(storage, dict):
            return {key: MemoryMappedTensorStorage._get(t, idx) for key, t in storage.items()}
        else:
            raise TypeError(f"Expected nested tuple/dict of tensors, found {type(storage)}.")

    def __getitem__(self, idx: int) -> NestedTensors:
        """Read the item stored at index `idx`."""
        return self._get(self._storage, idx)

    @staticmethod
    def _set(storage: NestedTensors, idx: int, data_point: NestedTensors) -> None:
        if isinstance(storage, torch.Tensor):
            assert isinstance(data_point, torch.Tensor)
            assert data_point.dtype is storage.dtype
            storage[idx] = data_point
        elif isinstance(storage, tuple):
            assert isinstance(data_point, tuple)
            assert len(data_point) == len(storage)
            for i in range(len(storage)):
                MemoryMappedTensorStorage._set(storage[i], idx, data_point[i])
        elif isinstance(storage, dict):
            assert isinstance(data_point, dict)
            assert set(data_point.keys()) == set(storage.keys())
            for key in storage:
                MemoryMappedTensorStorage._set(storage[key], idx, data_point[key])
        else:
            raise TypeError(f"Expected nested tuple/dict of tensors, found {type(storage)}.")

    def dump_dataset(self, ds):
        self._length = len(ds)
        self._storage = self._create_mmap_tensors(self._directory, ds[0], self._length)
        for idx in range(len(self)):
            self._set(self._storage, idx, ds[idx])


class FileTensorStorage(Storage):
    """A class implementing permanent storage of nested tensor datasets to disk as pickle files.

    This implements storage for `length` data points consisting of nested tensors of fixed types
    and shapes. `Storage` implements `__len__` and `__getitem__` and therefore can be used as a
    torch `Dataset`. To populate the storage, it also implements `dump_dataset`. It does _not_ keep
    track which slots have or have not been populated.

    `Storage` is given a path to a directory, where it creates (or accesses, if they already exist)
    pickle files one for each point in the dataset.

    Args:
        directory: Path to a directory.
    """

    def __init__(self, directory: str) -> None:
        super().__init__(directory)

    def dump_dataset(self, ds: torch.utils.data.Dataset) -> None:
        for i in range(len(ds)):
            torch.save(ds[i], self._compose_file_path_from_index(i))

    def __getitem__(self, idx: int) -> Any:
        if not hasattr(self, "_length"):
            self.load_dataset(None)
        return torch.load(self._compose_file_path_from_index(idx))

    def load_dataset(self, directory: Union[str, Path]):
        self._length = len([x for x in os.listdir(self._directory) if x.endswith(".pt")])

    def _compose_file_path_from_index(self, idx: int) -> str:
        return os.path.join(self._directory, f"{idx}.pt")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.shift.ks_detector import KolmogorovSmirnovCovariateShiftDetector
from renate.shift.mmd_detectors import MMDCovariateShiftDetector


@pytest.mark.parametrize(
    "detector",
    [
        MMDCovariateShiftDetector(feature_extractor=None, num_permutations=100),
        KolmogorovSmirnovCovariateShiftDetector(feature_extractor=None),
    ],
)
def test_shift_detector_identical_data(detector):
    """We expect low scores for identical data."""
    dataset = torch.utils.data.TensorDataset(torch.randn(100, 2))
    detector.fit(dataset)
    score = detector.score(dataset)
    assert score == 0.0


@pytest.mark.parametrize(
    "detector",
    [
        MMDCovariateShiftDetector(feature_extractor=None, num_permutations=100),
        KolmogorovSmirnovCovariateShiftDetector(feature_extractor=None),
    ],
)
def test_shift_detector_disjoint_data(detector):
    """We expect high scores for very different data (two disjoint Gaussian blobs)."""
    dataset_ref = torch.utils.data.TensorDataset(torch.randn(100, 2))
    dataset_query = torch.utils.data.TensorDataset(torch.randn(100, 2) + 2.0)
    detector.fit(dataset_ref)
    score = detector.score(dataset_query)
    assert score == 1.0



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest

from renate.utils.distributed_strategies import (
    _SUPPORTED_STRATEGIES,
    _UNSUPPORTED_STRATEGIES,
    create_strategy,
)


@pytest.mark.parametrize("devices", [1, 2, 3, 10])
@pytest.mark.parametrize("strategy_name", _SUPPORTED_STRATEGIES)
def test_valid_strategy_creation(devices, strategy_name):
    assert isinstance(create_strategy(devices, strategy_name), (str, object, None))


@pytest.mark.parametrize("devices", [1, 2, 3, 10])
@pytest.mark.parametrize("strategy_name", _UNSUPPORTED_STRATEGIES)
def test_invalid_strategy_creation(devices, strategy_name):
    with pytest.raises(ValueError) as _:
        create_strategy(devices, strategy_name)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
import os
from pathlib import Path
from typing import Callable, Optional, Tuple, Union

import pandas as pd
import torch
from torch.utils.data import Dataset, TensorDataset
from torchvision.datasets.utils import check_integrity

from renate import defaults
from renate.utils.file import download_folder_from_s3
from renate.utils.pytorch import randomly_split_data


class RenateDataModule(abc.ABC):
    """Data modules bundle code for data loading and preparation.

    A data module implements two methods for data preparation:
    - `prepare_data()` downloads the data to the local machine and unpacks it.
    - `setup()` creates pytorch dataset objects that return training, test and (possibly) validation
    data.
    These two steps are separated to streamline the process when launching multiple training jobs
    simultaneously, e.g., for hyperparameter optimization. In this case, `prepare_data()` is only
    called once per machine.

    After these two methods have been called, the data can be accessed using
    - `train_data()`
    - `test_data()`
    - `val_data()`,
    which return torch datasets (`torch.utils.data.Dataset`).

    Args:
        data_path: the path to the data to be loaded.
        src_bucket: the name of the s3 bucket.
        src_object_name: the folder path in the s3 bucket.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Union[Path, str, None] = None,
        src_object_name: Union[Path, str, None] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super().__init__()
        self._data_path = data_path
        self._src_bucket = src_bucket
        self._src_object_name = src_object_name
        self._train_data: Optional[Dataset] = None
        self._val_data: Optional[Dataset] = None
        self._test_data: Optional[Dataset] = None
        self._train_collate_fn: Optional[Callable] = None
        self._val_collate_fn: Optional[Callable] = None
        self._test_collate_fn: Optional[Callable] = None
        assert 0.0 <= val_size <= 1.0
        self._val_size = val_size
        self._seed = seed

        self._dataset_name: str = ""

    @abc.abstractmethod
    def prepare_data(self) -> None:
        """Downloads datasets."""
        pass

    @abc.abstractmethod
    def setup(self) -> None:
        """Set up train, test and val datasets."""
        pass

    def train_data(self) -> Dataset:
        """Returns training dataset."""
        return self._train_data

    def val_data(self) -> Dataset:
        """Returns validation dataset."""
        return self._val_data

    def test_data(self) -> Dataset:
        """Returns test dataset."""
        return self._test_data

    def train_collate_fn(self) -> Optional[Callable]:
        """Returns collate_fn for train DataLoader."""
        return self._train_collate_fn

    def val_collate_fn(self) -> Optional[Callable]:
        """Returns collate_fn for validation DataLoader."""
        return self._val_collate_fn

    def test_collate_fn(self) -> Optional[Callable]:
        """Returns collate_fn for test DataLoader."""
        return self._test_collate_fn

    def _verify_file(self, file_name: str) -> bool:
        """A helper function that verifies that the required dataset files are downloaded and
        correct.
        """
        return check_integrity(
            os.path.join(self._data_path, self._dataset_name, file_name),
            self.md5s[file_name],
        )

    def _split_train_val_data(self, train_data: Dataset) -> Tuple[Dataset, Dataset]:
        """A helper function that splits the train data into train and validation sets."""
        if self._val_size == 0.0:
            return train_data, None
        else:
            return randomly_split_data(
                train_data, [1.0 - self._val_size, self._val_size], self._seed
            )


class CSVDataModule(RenateDataModule):
    """A data module loading data from CSV files.

    Args:
        data_path: Path to the folder containing the files.
        train_filename: Name of the CSV file containing the training data.
        test_filename: Name of the CSV file containing the test data.
        src_bucket: Name of an s3 bucket. If specified, the folder given by `src_object_name` will
            be downloaded from S3 to `data_path`.
        src_object_name: Folder path in the s3 bucket.
        target_name: the header of the column containing the target values.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_path: Union[Path, str],
        train_filename: Union[Path, str] = "train.csv",
        test_filename: Union[Path, str] = "test.csv",
        target_name: str = "y",
        src_bucket: Union[Path, str, None] = None,
        src_object_name: Union[Path, str, None] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super(CSVDataModule, self).__init__(
            data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self._train_filename = train_filename
        self._test_filename = test_filename
        self._target_name = target_name

    def prepare_data(self) -> None:
        """Downloads data folder from S3 if applicable."""
        if self._src_bucket is not None:
            download_folder_from_s3(self._src_bucket, self._src_object_name, self._data_path)

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        X, y = self._process_csv_data(str(self._train_filename))
        train_data = TensorDataset(X, y)
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        X, y = self._process_csv_data(str(self._test_filename))
        self._test_data = TensorDataset(X, y)

    def _process_csv_data(self, filename: str) -> Tuple[torch.Tensor, torch.Tensor]:
        """Reads data from a CSV file and returns features and labels."""
        data_path = os.path.join(self._data_path, filename)
        data = pd.read_csv(data_path)
        if self._target_name not in data.columns:
            raise KeyError(f"{self._target_name} is not a valid target name.")
        y = torch.from_numpy(data[self._target_name].to_numpy())
        data = data.loc[:, data.columns != self._target_name]
        X = torch.from_numpy(data.to_numpy()).float()
        return X, y



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset

from renate import defaults
from renate.memory import InfiniteBuffer
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.learner import Learner
from renate.updaters.model_updater import SingleTrainingLoopUpdater


class JointLearner(Learner):
    """A Learner that implements the Joint strategy.

    This is a simple strategy that trains the model on all previously observed data.
    Each time a new chunk of data is observed, the model is reinitialized and retrained on
    all the previously observed data and the new chunk of data. The buffer holding the previous
    data is updated with the new chunk of data.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._memory_buffer = InfiniteBuffer(
            transform=self._train_transform,
            target_transform=self._train_target_transform,
        )

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        super().on_save_checkpoint(checkpoint=checkpoint)
        checkpoint["memory_buffer"] = self._memory_buffer.state_dict()

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        super().on_load_checkpoint(checkpoint)
        self._memory_buffer.load_state_dict(checkpoint["memory_buffer"])

    def save(self, output_state_dir: str) -> None:
        super().save(output_state_dir)
        buffer_dir = os.path.join(output_state_dir, "memory_buffer")
        os.makedirs(buffer_dir, exist_ok=True)
        self._memory_buffer.save(buffer_dir)

    def load(self, input_state_dir: str) -> None:
        super().load(input_state_dir)
        self._memory_buffer.load(os.path.join(input_state_dir, "memory_buffer"))

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        """Called before a model update starts."""
        super().on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._memory_buffer.update(train_dataset)

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self._memory_buffer,
            batch_size=self._batch_size,
            shuffle=True,
            generator=self._rng,
            pin_memory=True,
            collate_fn=self._train_collate_fn,
        )

    def training_step(
        self,
        batch: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        batch_idx: int,
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        batch, _ = batch
        return super().training_step(batch=batch, batch_idx=batch_idx)


class JointModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=JointLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import toml
from typing import List
import shutil


def generate_requirements_file_for_sagemaker_training_jobs(
    toml_file: str, keys: List[str], output_path: str
):
    dependencies = []

    # create a copy of the main requirements.txt file
    shutil.copyfile("./requirements.txt", "test/integration_tests/requirements.txt")

    # Parse the .toml file
    with open(toml_file, "r") as file:
        pyproject_toml = toml.load(file)

    for key in keys:
        value = pyproject_toml["project"]["optional-dependencies"][key]
        dependencies += value

    # Write the values into the output file
    with open(f"{output_path}/requirements.txt", "a") as file:
        for dependency in dependencies:
            file.write(f"{dependency}\n")


if __name__ == "__main__":
    generate_requirements_file_for_sagemaker_training_jobs(
        "./pyproject.toml", ["avalanche", "benchmark", "dev"], "test/integration_tests"
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
import os
from collections import defaultdict

import pytest
import torch

from renate.data.datasets import NestedTensorDataset
from renate.memory.buffer import (
    GreedyClassBalancingBuffer,
    InfiniteBuffer,
    ReservoirBuffer,
    SlidingWindowBuffer,
)


def nested_tensors_equal(t1, t2):
    if type(t1) is not type(t2):
        print("Type mismatch")
        return False
    if isinstance(t1, torch.Tensor):
        return torch.allclose(t1, t2, rtol=1e-3)
    if isinstance(t1, tuple):
        if len(t1) != len(t2):
            print("Tuple length mismatch")
            return False
        return all(nested_tensors_equal(t1_, t2_) for t1_, t2_ in zip(t1, t2))
    if isinstance(t1, dict):
        if set(t1.keys()) != set(t2.keys()):
            print("Dict key mismatch")
            return False
        return all(nested_tensors_equal(t1[key], t2[key]) for key in t1.keys())


@pytest.mark.parametrize("buffer_cls", [ReservoirBuffer, SlidingWindowBuffer])
@pytest.mark.parametrize("max_size", [1, 10, 100])
@pytest.mark.parametrize("num_updates", [1, 10])
@pytest.mark.parametrize(
    "dataset",
    [
        torch.utils.data.TensorDataset(torch.randn(10, 2)),
        torch.utils.data.TensorDataset(torch.randn(100, 2)),
        torch.utils.data.TensorDataset(torch.randn(10, 2), torch.arange(10)),
        torch.utils.data.TensorDataset(torch.randn(100, 2), torch.arange(100)),
        NestedTensorDataset({"X": torch.randn(10, 2), "y": torch.arange(10)}),
        NestedTensorDataset({"X": torch.randn(100, 2), "y": torch.arange(100)}),
        NestedTensorDataset(({"X": torch.randn(10, 2), "z": torch.randn(10)}, torch.arange(10))),
        NestedTensorDataset(({"X": torch.randn(100, 2), "z": torch.randn(100)}, torch.arange(100))),
    ],
)
def test_buffer_respects_max_size(buffer_cls, max_size, num_updates, dataset):
    buffer = buffer_cls(max_size)
    assert len(buffer) == 0
    for i in range(num_updates):
        buffer.update(dataset)
        assert len(buffer) == min((i + 1) * len(dataset), max_size)


@pytest.mark.parametrize("buffer", [ReservoirBuffer(30), SlidingWindowBuffer(30)])
def test_batching_with_metadata(buffer):
    X = torch.randn(100, 3)
    y = torch.randint(0, 10, size=(100,))
    ds = torch.utils.data.TensorDataset(X, y)
    metadata = {"foo": torch.ones(100)}
    buffer.update(ds, metadata)

    dataloader = torch.utils.data.DataLoader(buffer, batch_size=3)
    for batch, metadata_batch in dataloader:
        assert len(batch) == 2
        assert isinstance(metadata_batch, dict)
        assert "foo" in metadata_batch
        assert isinstance(metadata_batch["foo"], torch.Tensor)
        assert torch.all(metadata_batch["foo"] == 1.0)


@pytest.mark.parametrize("buffer", [ReservoirBuffer(30), SlidingWindowBuffer(30)])
def test_get_and_set_metadata(buffer):
    X = torch.randn(100, 3)
    y = torch.randint(0, 10, size=(100,))
    ds = torch.utils.data.TensorDataset(X, y)
    metadata = {"foo": torch.ones(100)}
    buffer.update(ds, metadata)

    foo_values = buffer.get_metadata("foo")
    assert isinstance(foo_values, torch.Tensor)
    assert foo_values.size() == (len(buffer),)
    assert torch.all(foo_values == 1.0)

    buffer.set_metadata("foo", 2 * torch.ones(len(buffer)))
    foo_values = buffer.get_metadata("foo")
    assert isinstance(foo_values, torch.Tensor)
    assert foo_values.size() == (len(buffer),)
    assert torch.all(foo_values == 2.0)


@pytest.mark.parametrize("buffer", [ReservoirBuffer(30), SlidingWindowBuffer(30)])
def test_getitem_returns_points_and_metadata(buffer):
    X = torch.randn(100, 3)
    y = torch.randint(0, 10, size=(100,))
    ds = torch.utils.data.TensorDataset(X, y)
    metadata = {"foo": torch.ones(100)}
    buffer.update(ds, metadata)

    for idx in [0, 1, 10, 29]:
        datapoint, metadata = buffer[idx]
        assert isinstance(metadata, dict) and ("foo" in metadata)
        assert metadata["foo"] == 1.0
        assert isinstance(datapoint, tuple) and len(datapoint) == 2
        assert datapoint[0].size() == (3,)
        assert datapoint[1].size() == ()

    for idx in [30, 31, 50]:
        with pytest.raises(IndexError):
            datapoint[idx]


def test_reservoir_is_uniform():
    """Present buffer with numbers from 1-10k. It should subsample uniformly, resulting in a mean of
    5k with a standard deviation of ~91.
    """
    buffer = ReservoirBuffer(max_size=1000)
    for i in range(10):
        x = torch.arange(i * 1000, (i + 1) * 1000, dtype=torch.float) + 1
        ds = torch.utils.data.TensorDataset(x)
        buffer.update(ds)
    xs = torch.stack([buffer[i][0][0] for i in range(len(buffer))], dim=0)
    assert 5000 - 5 * 91 < xs.mean() < 5000 + 5 * 91


def test_greedy_is_balanced(tmpdir):
    """Present buffer with 10 classes with respect to 10000 samples, where each class has 1000
    samples. When inspecting the ._class_counts attribute, it should be uniform.
    """
    buffer = GreedyClassBalancingBuffer(max_size=100)
    X = torch.ones(10000, 1)
    Y = torch.ones(10000, 1, dtype=torch.long)
    for i in range(10):
        Y[i * 1000 : (i + 1) * 1000] = i
    # Randomise the Y such that the classes are not presented in a sequence
    Y = Y[torch.randperm(10000)]
    # Split the data into 10 chunks and perform 10 updates
    for i in range(10):
        ds = torch.utils.data.TensorDataset(
            X[i * 1000 : (i + 1) * 1000], Y[i * 1000 : (i + 1) * 1000]
        )
        buffer.update(ds)
        buffer.save(tmpdir)
        state_dict = buffer.state_dict()
        del buffer
        buffer = GreedyClassBalancingBuffer(max_size=100)
        buffer.load_state_dict(state_dict)
        buffer.load(tmpdir)
    counts = torch.tensor([len(v) for v in buffer._indices_by_class.values()], dtype=torch.float32)
    label_counts = defaultdict(int)
    assert torch.all(counts >= 10 - 1)
    assert torch.all(counts <= 10 + 1)
    for i in range(len(buffer)):
        (_, y), _ = buffer[i]
        label_counts[y.item()] += 1
    buffer_class_counts = {k: len(v) for k, v in buffer._indices_by_class.items()}
    assert buffer_class_counts == label_counts


@pytest.mark.parametrize("max_size", [1, 10, 100])
@pytest.mark.parametrize("num_batches", [1, 10])
@pytest.mark.parametrize("batch_size", [1, 10, 20])
def test_sliding_window_keeps_most_recent(max_size, num_batches, batch_size):
    buffer = SlidingWindowBuffer(max_size)
    for i in range(num_batches):
        ds = torch.utils.data.TensorDataset(torch.arange(i * batch_size, (i + 1) * batch_size))
        buffer.update(ds)
    xs = torch.stack([buffer[i][0][0] for i in range(len(buffer))], dim=0)
    assert xs.max() == num_batches * batch_size - 1
    assert xs.min() == max(num_batches * batch_size - max_size, 0)


@pytest.mark.parametrize(
    "state_dict",
    [
        {
            "max_size": None,
            "storage_mode": "in_memory",
            "seed": 23,
            "count": 23,
            "data_points": [],
            "metadata": [],
        },
        {
            "max_size": 5,
            "storage_mode": None,
            "seed": 23,
            "count": 23,
            "data_points": [(torch.ones((5,)),) for _ in range(6)],
            "metadata": [],
        },
        {
            "max_size": 5,
            "storage_mode": "in_memory",
            "seed": "1",
            "count": 23,
            "data_points": [(torch.ones((5,)),) for _ in range(6)],
            "metadata": [],
        },
        {
            "max_size": 5,
            "storage_mode": "unknown_storage_mode",
            "seed": 23,
            "count": None,
            "data_points": [(torch.ones((5,)),) for _ in range(6)],
            "metadata": [(torch.ones((5,)),) for _ in range(6)],
        },
        {
            "max_size": 5,
            "storage_mode": "unknown_storage_mode",
            "seed": 23,
            "count": 10,
            "data_points": torch.ones(
                5,
            ),
            "metadata": [(torch.ones((5,)),) for _ in range(6)],
        },
        {
            "max_size": 5,
            "storage_mode": "unknown_storage_mode",
            "seed": 23,
            "count": 23,
            "data_points": [(torch.ones((5,)),) for _ in range(6)],
            "metadata": torch.ones((5,)),
        },
    ],
)
def test_buffer_loading_invalid_state_dict(state_dict):
    buffer = SlidingWindowBuffer(max_size=10)
    with pytest.raises(Exception):
        buffer.load_state_dict(state_dict)


def test_infinite_buffer():
    buffer = InfiniteBuffer()
    for i in range(5):
        sample_dataset = torch.utils.data.TensorDataset(torch.ones((10, 3)) * i)
        buffer.update(sample_dataset)

    for i in range(5):
        for j in range(10):
            assert torch.all(torch.eq(buffer[i * 10 + j][0][0], torch.ones((3)) * i))


@pytest.mark.parametrize("max_size", [1, 10, 100])
@pytest.mark.parametrize("buffer_cls", [ReservoirBuffer, SlidingWindowBuffer])
def test_buffer_same_size_on_disk_after_updates(tmpdir, max_size, buffer_cls):
    """Test that the buffer size on disk is the same after consequtive updates
    given that the max_size is reached."""
    buffer = buffer_cls(max_size=max_size)
    for i in range(20):
        ds = torch.utils.data.TensorDataset(torch.arange(i * 10, (i + 1) * 10))
        buffer.update(ds)

    torch.save(buffer.state_dict(), os.path.join(tmpdir, "buffer.pt"))
    disk_size = os.path.getsize(os.path.join(tmpdir, "buffer.pt"))
    state_dict = torch.load(os.path.join(tmpdir, "buffer.pt"))
    buffer.load_state_dict(state_dict)
    for i in range(20):
        ds = torch.utils.data.TensorDataset(torch.arange(i * 10, (i + 1) * 10))
        buffer.update(ds)

    torch.save(buffer.state_dict(), os.path.join(tmpdir, "buffer.pt"))
    if not isinstance(buffer, InfiniteBuffer):
        assert disk_size == os.path.getsize(os.path.join(tmpdir, "buffer.pt"))
    else:
        assert disk_size < os.path.getsize(os.path.join(tmpdir, "buffer.pt"))


@pytest.mark.parametrize("buffer_cls", [ReservoirBuffer, SlidingWindowBuffer])
@pytest.mark.parametrize("max_size", [10, 100])
@pytest.mark.parametrize("num_updates", [0, 1, 2])
@pytest.mark.parametrize(
    "dataset",
    [
        torch.utils.data.TensorDataset(torch.randn(10, 2)),
        torch.utils.data.TensorDataset(torch.randn(10, 2), torch.arange(10)),
        NestedTensorDataset({"X": torch.randn(10, 2), "y": torch.arange(10)}),
        NestedTensorDataset(({"X": torch.randn(10, 2), "z": torch.randn(10)}, torch.arange(10))),
    ],
)
@pytest.mark.parametrize("metadata", [None, {"a": torch.arange(10), "b": torch.zeros(10, 2)}])
def test_load_and_save_buffer(tmpdir, buffer_cls, max_size, num_updates, dataset, metadata):
    """Tests loading an saving of the buffer state."""
    buffer = buffer_cls(max_size)
    for _ in range(2):
        for _ in range(num_updates):
            buffer.update(dataset, metadata)
        elements_before = [copy.deepcopy(buffer[i]) for i in range(len(buffer))]
        buffer.save(tmpdir)
        state_dict = buffer.state_dict()
        del buffer
        buffer = buffer_cls(max_size)
        buffer.load_state_dict(state_dict)
        buffer.load(tmpdir)
        for j in range(len(buffer)):
            assert nested_tensors_equal(buffer[j], elements_before[j])


@pytest.mark.parametrize("buffer_cls", [ReservoirBuffer, SlidingWindowBuffer])
@pytest.mark.parametrize("max_size", [10, 100])
@pytest.mark.parametrize("num_updates", [0, 1, 2])
@pytest.mark.parametrize(
    "dataset",
    [
        torch.utils.data.TensorDataset(torch.randn(10, 2), torch.arange(10)),
    ],
)
@pytest.mark.parametrize("metadata", [None, {"a": torch.arange(10), "b": torch.zeros(10, 2)}])
@pytest.mark.parametrize("buffer_transform", [None, lambda x: x * 2])
@pytest.mark.parametrize("buffer_target_transform", [None, lambda y: y // 5])
def test_load_and_save_buffer_with_transforms(
    tmpdir,
    buffer_cls,
    max_size,
    num_updates,
    dataset,
    metadata,
    buffer_transform,
    buffer_target_transform,
):
    """Tests loading an saving of the buffer state."""
    buffer = buffer_cls(
        max_size, transform=buffer_transform, target_transform=buffer_target_transform
    )
    for _ in range(2):
        for _ in range(num_updates):
            buffer.update(dataset, metadata)
        elements_before = [copy.deepcopy(buffer[i]) for i in range(len(buffer))]
        buffer.save(tmpdir)
        state_dict = buffer.state_dict()
        del buffer
        buffer = buffer_cls(
            max_size, transform=buffer_transform, target_transform=buffer_target_transform
        )
        buffer.load_state_dict(state_dict)
        buffer.load(tmpdir)
        for j in range(len(buffer)):
            assert nested_tensors_equal(buffer[j], elements_before[j])



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path
from typing import Callable, Dict, IO, Optional, Union

import torch
from avalanche.training.plugins.checkpoint import CheckpointPlugin, FileSystemCheckpointStorage

from renate import defaults


class RenateFileSystemCheckpointStorage(FileSystemCheckpointStorage):
    def _make_checkpoint_dir(self, checkpoint_name: str) -> Path:
        return self.directory

    def _make_checkpoint_file_path(self, checkpoint_name: str) -> Path:
        return Path(defaults.learner_state_file(str(self._make_checkpoint_dir(checkpoint_name))))

    def checkpoint_exists(self, checkpoint_name: str) -> bool:
        return self._make_checkpoint_file_path(checkpoint_name).exists()

    def store_checkpoint(
        self, checkpoint_name: str, checkpoint_writer: Callable[[IO[bytes]], None]
    ):
        checkpoint_file = self._make_checkpoint_file_path(checkpoint_name)
        if checkpoint_file.exists():
            checkpoint_file.unlink()
        super().store_checkpoint(
            checkpoint_name=checkpoint_name, checkpoint_writer=checkpoint_writer
        )


class RenateCheckpointPlugin(CheckpointPlugin):
    def __init__(
        self,
        storage: RenateFileSystemCheckpointStorage,
        map_location: Optional[Union[str, torch.device, Dict[str, str]]] = None,
    ):
        super().__init__(storage=storage, map_location=map_location)

    def load_checkpoint_if_exists(self):
        if not self.storage.checkpoint_exists(defaults.LEARNER_CHECKPOINT_NAME):
            return None, 0
        loaded_checkpoint = self.storage.load_checkpoint(
            defaults.LEARNER_CHECKPOINT_NAME, self.load_checkpoint
        )
        return loaded_checkpoint["strategy"], loaded_checkpoint["exp_counter"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from typing import Union

import numpy as np
import numpy.typing as npt
import torch
import torch.nn as nn
from sklearn.cluster import KMeans


class TaskEstimator(nn.Module, ABC):
    """An ABC that all task estimator methods inherit.

    They implement two methods `update_task_prototypes` and `infer_task`.
    """

    @abstractmethod
    def update_task_prototypes(self):
        return

    @abstractmethod
    def infer_task(self):
        return


class TaskPrototypes(TaskEstimator):
    """Task identification method proposed in S-Prompts.

    Args:
        task_id: The current update id of the method. Required to deserialize.
        clusters_per_task: Number of clusters to use in K-means.
        embedding_size: Embedding size of the transformer features.
    """

    def __init__(self, task_id, clusters_per_task, embedding_size) -> None:
        super().__init__()
        self.register_buffer(
            "_training_feat_centroids",
            torch.empty(task_id * clusters_per_task, embedding_size),
        )
        self.register_buffer(
            "_training_feat_task_ids",
            torch.full(
                (self._training_feat_centroids.size(0),), fill_value=task_id, dtype=torch.long
            ),
        )
        self._clusters_per_task = clusters_per_task
        self._task_id = task_id
        self._embedding_size = embedding_size

    @torch.no_grad()
    def update_task_prototypes(
        self,
        features: Union[torch.Tensor, npt.ArrayLike],
        labels: Union[torch.Tensor, npt.ArrayLike],
    ) -> None:
        # At training.
        if isinstance(features, torch.Tensor):
            features = features.cpu().numpy()

        # l2 normalize features:
        features = features / np.power(np.einsum("ij, ij -> i", features, features), 0.5)[:, None]

        centroids = torch.from_numpy(
            KMeans(n_clusters=self._clusters_per_task, random_state=0)
            .fit(features)
            .cluster_centers_
        ).to(self._training_feat_centroids.device)

        self._training_feat_centroids = torch.cat(
            [
                self._training_feat_centroids,
                centroids,
            ]
        )
        self._training_feat_task_ids = torch.cat(
            [
                self._training_feat_task_ids,
                torch.full(
                    (centroids.size(0),),
                    fill_value=self._task_id,
                    dtype=torch.int8,
                    device=self._training_feat_task_ids.device,
                ),
            ]
        )

    def infer_task(self, features: torch.Tensor) -> torch.Tensor:
        # At inference.
        if self._training_feat_centroids.numel() > 0:
            features = torch.nn.functional.normalize(features)
            nearest_p_inds = torch.cdist(features, self._training_feat_centroids, p=2).argmin(1)
            return self._training_feat_task_ids[nearest_p_inds]
        else:
            return None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import pytest
import torch
from avalanche.models import TrainEvalModel
from avalanche.training import ICaRLLossPlugin
from avalanche.training.plugins import EWCPlugin, LwFPlugin
from torch.optim import Adam, SGD

from conftest import (
    AVALANCHE_LEARNERS,
    AVALANCHE_LEARNER_HYPERPARAMETER_UPDATES,
    AVALANCHE_LEARNER_KWARGS,
)
from renate.updaters.avalanche.learner import (
    AvalancheEWCLearner,
    AvalancheICaRLLearner,
    AvalancheLwFLearner,
)
from renate.utils.avalanche import plugin_by_class


def check_learner_settings(
    learner,
    learner_kwargs,
    avalanche_learner,
    expected_model,
    expected_optimizer,
    expected_max_epochs,
    expected_device,
    expected_eval_every,
    expected_batch_size,
    expected_loss_fn=None,
):
    if isinstance(learner, AvalancheICaRLLearner):
        assert isinstance(avalanche_learner._criterion, ICaRLLossPlugin)
        assert avalanche_learner.eval_every == -1
        assert isinstance(avalanche_learner.model, TrainEvalModel)
        assert avalanche_learner.model.feature_extractor == expected_model.get_backbone()
        assert avalanche_learner.model.classifier == expected_model.get_predictor()
    else:
        assert avalanche_learner.eval_every == expected_eval_every
        assert avalanche_learner.model == expected_model
        if expected_loss_fn is None:
            assert avalanche_learner._criterion == learner_kwargs["loss_fn"]
        else:
            assert avalanche_learner._criterion == expected_loss_fn
    assert avalanche_learner.optimizer == expected_optimizer
    assert avalanche_learner.train_epochs == expected_max_epochs
    assert avalanche_learner.train_mb_size == expected_batch_size
    assert avalanche_learner.eval_mb_size == learner_kwargs["batch_size"]

    assert avalanche_learner.device == expected_device
    if isinstance(learner, AvalancheEWCLearner):
        assert (
            plugin_by_class(EWCPlugin, avalanche_learner.plugins).ewc_lambda
            == learner_kwargs["ewc_lambda"]
        )
    elif isinstance(learner, AvalancheLwFLearner):
        assert (
            plugin_by_class(LwFPlugin, avalanche_learner.plugins).lwf.alpha
            == learner_kwargs["alpha"]
        )
        assert (
            plugin_by_class(LwFPlugin, avalanche_learner.plugins).lwf.temperature
            == learner_kwargs["temperature"]
        )


@pytest.mark.parametrize("learner_class", AVALANCHE_LEARNERS)
def test_update_settings(learner_class):
    """Test settings after Avalanche wrappers creation and update."""
    learner_kwargs = AVALANCHE_LEARNER_KWARGS[learner_class]
    expected_model = pytest.helpers.get_renate_module_mlp(
        num_inputs=1,
        num_outputs=1,
        hidden_size=1,
        num_hidden_layers=1,
        add_icarl_class_means=learner_class == AvalancheICaRLLearner,
    )
    plugins = []
    expected_max_epochs = 10
    expected_loss_fn = pytest.helpers.get_loss_fn("mean")
    expected_optimizer = SGD(expected_model.parameters(), lr=0.1)
    expected_device = torch.device("cpu")
    expected_eval_every = -1
    expected_batch_size = learner_kwargs["batch_size"]
    if "batch_memory_frac" in learner_kwargs:
        expected_batch_size = expected_batch_size - int(
            learner_kwargs["batch_memory_frac"] * expected_batch_size
        )
    learner = learner_class(
        model=expected_model,
        optimizer=None,
        loss_fn=expected_loss_fn,
        **learner_kwargs,
    )
    avalanche_learner = learner.create_avalanche_learner(
        plugins=plugins,
        optimizer=expected_optimizer,
        train_epochs=expected_max_epochs,
        device=expected_device,
        eval_every=expected_eval_every,
    )
    check_learner_settings(
        learner,
        learner_kwargs,
        avalanche_learner,
        expected_model=expected_model,
        expected_loss_fn=expected_loss_fn,
        expected_optimizer=expected_optimizer,
        expected_max_epochs=expected_max_epochs,
        expected_device=expected_device,
        expected_eval_every=expected_eval_every,
        expected_batch_size=expected_batch_size,
    )

    # Update
    expected_max_epochs = 20
    expected_optimizer = Adam(expected_model.parameters(), lr=0.3)
    expected_device = torch.device("cuda")
    expected_eval_every = 1
    for key, value in AVALANCHE_LEARNER_HYPERPARAMETER_UPDATES[learner_class].items():
        setattr(learner, f"_{key}", value)
    learner.update_settings(
        avalanche_learner=avalanche_learner,
        plugins=plugins,
        optimizer=expected_optimizer,
        max_epochs=expected_max_epochs,
        device=expected_device,
        eval_every=expected_eval_every,
    )
    learner_kwargs.update(AVALANCHE_LEARNER_HYPERPARAMETER_UPDATES[learner_class])
    check_learner_settings(
        learner,
        learner_kwargs,
        avalanche_learner,
        expected_model=expected_model,
        expected_loss_fn=expected_loss_fn,
        expected_optimizer=expected_optimizer,
        expected_max_epochs=expected_max_epochs,
        expected_device=expected_device,
        expected_eval_every=expected_eval_every,
        expected_batch_size=expected_batch_size,
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import torch

from renate.models.renate_module import RenateModule
from renate.updaters.learner_components.component import Component
from renate.utils.pytorch import reinitialize_model_parameters


class ReinitializationComponent(Component):
    """Resets the model using each layer's built-in reinitialization logic.

    See also `renate.utils.torch_utils.reinitialize_model_parameters`.
    """

    def on_train_start(self, model: RenateModule) -> None:
        reinitialize_model_parameters(model)


class ShrinkAndPerturbReinitializationComponent(Component):
    """Shrinking and Perturbation reinitialization through scaling the
    weights and adding random noise.

    Ash, J., & Adams, R. P. (2020). On warm-starting neural network training.
    Advances in Neural Information Processing Systems, 33, 3884-3894.

    Args:
        shrink_factor: A scaling coefficient applied to shrink the weights.
        sigma: variance of the random Gaussian noise added to the weights.
    """

    def __init__(self, shrink_factor: float, sigma: float) -> None:
        self._shrink_factor = shrink_factor
        self._sigma = sigma
        super().__init__()

    def _verify_attributes(self) -> None:
        """Verify if attributes have valid values."""
        super()._verify_attributes()
        assert self._shrink_factor > 0.0, "Shrink factor must be positive."
        assert self._sigma >= 0, "Sigma must be non-negative."

    @torch.no_grad()
    def on_train_start(self, model: RenateModule) -> None:
        """Shrink and perturb the model's weights."""
        for p in model.parameters():
            if self._shrink_factor != 1.0:
                p.mul_(self._shrink_factor)
            if self._sigma != 0.0:
                p.add_(self._sigma * torch.randn(p.size(), device=p.device))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

import torch

from renate.shift.detector import ShiftDetectorWithFeatureExtractor
from renate.shift.kernels import RBFKernel
from renate.shift.mmd_helpers import mmd


class MMDCovariateShiftDetector(ShiftDetectorWithFeatureExtractor):
    """A kernel maximum mean discrepancy (MMD) test.

    This test was proposed by

    [1] Gretton, A., et al. A kernel two-sample test. JMLR (2012).

    We currently do not expose the choice of kernel. It defaults to an RBF kernel with a lengthscale
    set via the median heuristic.

    The detector computes an approximate p-value via a permutation test. The `score` method returns
    `1 - p_value` to conform to the convention that high scores indicate a shift.

    Args:
        feature_extractor: A pytorch model used as feature extractor.
        num_permutations: Number of permutations for permutation test.
        batch_size: Batch size used to iterate over datasets.
        num_preprocessing_workers: Number of workers used in data loaders.
        device: Device to use for computations inside the detector.
    """

    def __init__(
        self,
        feature_extractor: Optional[torch.nn.Module] = None,
        num_permutations: int = 1000,
        batch_size: int = 32,
        num_preprocessing_workers: int = 0,
        device: str = "cpu",
    ) -> None:
        super().__init__(feature_extractor, batch_size, num_preprocessing_workers, device)
        self._num_permutations = num_permutations

    def _fit_with_features(self, X: torch.Tensor):
        self._X_ref = X

    def _score_with_features(self, X: torch.Tensor) -> float:
        _, p_val = mmd(self._X_ref, X, kernel=RBFKernel(), num_permutations=self._num_permutations)
        return 1.0 - p_val.item()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch
from transformers import AutoTokenizer

from renate.utils.hf_utils import BatchEncoding, DataCollatorWithPaddingForWildTime


@pytest.mark.parametrize(
    "features, shape, error",
    [
        (
            [
                ({"input_ids": [0, 1, 2]}, 0),
                ({"input_ids": [0, 1, 2, 3, 4, 5]}, 1),
            ],
            torch.Size([2, 6]),
            False,
        ),
        (
            [
                (({"input_ids": [1, 2, 4]}, 0), {}),
                (({"input_ids": [4, 5, 6, 7, 8, 9]}, 1), {}),
            ],
            torch.Size([2, 6]),
            False,
        ),
        (
            [
                (({"input_ids": [1, 2, 4]}, 0), {"logits": [0.1, 0.2, 0.3]}),
                (({"input_ids": [4, 5, 6, 7, 8, 9]}, 1), {"logits": [0.3, 0.4, 0.5]}),
            ],
            torch.Size([2, 6]),
            False,
        ),
        (
            [[1, ({"input_ids": [0, 1, 2]}, 0)], [2, ({"input_ids": [0, 1, 2, 3, 4, 5]}, 1)]],
            torch.Size([2]),
            False,
        ),
        (
            [
                ({"input_ids": [1, 2, 4]}, 0, {}),
            ],
            torch.Size([1, 3]),
            True,
        ),
    ],
)
@pytest.mark.parametrize("tokenizer", [AutoTokenizer.from_pretrained("bert-base-cased")])
def test_serialize_wildtime_collator(features, shape, error, tokenizer):
    if not error:
        collator = DataCollatorWithPaddingForWildTime(tokenizer)
        out = collator(features)[0]
        print(collator(features))
        if isinstance(out, (dict, BatchEncoding)):
            assert out["input_ids"].shape == shape
        elif isinstance(out, torch.Tensor):
            assert out.shape == shape
        else:
            assert out[0]["input_ids"].shape == shape
    else:
        collator = DataCollatorWithPaddingForWildTime(tokenizer)
        with pytest.raises(Exception):
            out = collator(features)[0]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple

import torch
from torch.nn import Parameter
from torch.optim import Optimizer, SGD
from torch.optim.lr_scheduler import StepLR, _LRScheduler
from torchmetrics import Accuracy

from dummy_datasets import DummyTorchVisionDataModule
from renate.benchmark.models.mlp import MultiLayerPerceptron
from renate.data.data_module import RenateDataModule
from renate.models import RenateModule


def model_fn(model_state_url: Optional[str] = None) -> RenateModule:
    if model_state_url is None:
        return MultiLayerPerceptron(5 * 5, 10, 0, 64)
    state_dict = torch.load(model_state_url)
    return MultiLayerPerceptron.from_state_dict(state_dict)


def data_module_fn(
    data_path: str,
    val_size: float = 0.0,
    seed: int = 0,
) -> RenateDataModule:
    return DummyTorchVisionDataModule(transform=None, val_size=val_size, seed=seed)


def loss_fn(updater: Optional[str] = None) -> torch.nn.Module:
    if updater.startswith("Avalanche-"):
        return torch.nn.CrossEntropyLoss()
    return torch.nn.CrossEntropyLoss(reduction="none")


def optimizer_fn() -> Callable[[List[Parameter]], Optimizer]:
    return partial(SGD, lr=0.01)


def lr_scheduler_fn() -> Tuple[Callable[[Optimizer], _LRScheduler], str]:
    return partial(StepLR, step_size=10, gamma=0.1), "epoch"


def metrics_fn() -> Dict:
    return {"accuracy": Accuracy(task="multiclass", num_classes=10)}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import time
from typing import Any, Dict, Union

from syne_tune.backend.trial_status import Trial
from syne_tune.config_space import Domain, from_dict, to_dict
from syne_tune.experiments import ExperimentResult
from syne_tune.tuner_callback import TunerCallback
from tabulate import tabulate

logger = logging.getLogger(__name__)


class TuningLoggerCallback(TunerCallback):
    """Syne Tune Logging Callback when running hyperparameter optimization.

    Will report whenever improved results are obtained. If no better configuration is received
    within the last ten minutes, it will send a message to indicate it is still running.
    """

    def __init__(self, mode: str, metric: str):
        self._mode = mode
        self._metric = metric
        self._best_score = float("inf") * (1 if self._mode == "min" else -1)
        self._last_log = time.time()

    def _log(self, message: str) -> None:
        """Helper function to log a message."""
        logger.info(message)
        self._last_log = time.time()

    def on_trial_result(self, trial: Trial, status: str, result: Dict, decision: str) -> None:
        """Called whenever the tuner receives a new (intermediate) observation."""
        if self._mode == "min":
            is_better_config = result[self._metric] < self._best_score
            self._best_score = min(result[self._metric], self._best_score)
        else:
            is_better_config = result[self._metric] > self._best_score
            self._best_score = max(result[self._metric], self._best_score)
        if is_better_config:
            self._log(f"Metric `{self._metric}` improved to {self._best_score}.")
        if self._last_log < time.time() - 600:
            self._log("No better configuration found since last update. Continue searching...")


class TrainingLoggerCallback(TunerCallback):
    """Syne Tune Logging Callback when training a single configuration.

    Will report all metrics after each epoch."""

    def on_trial_result(self, trial: Trial, status: str, result: Dict, decision: str) -> None:
        """Called whenever the tuner receives a new (intermediate) observation."""
        result_table = tabulate(
            [[k, v] for k, v in result.items() if k.startswith("train_") or k.startswith("val_")]
        )
        logger.info(f"Epoch {result['epoch']}/{trial.config['max_epochs']}\n{result_table}")


def redirect_to_tmp(uri: str) -> str:
    """Changes uri in /opt/ml to /tmp.

    Syne Tune stores checkpoints by default in /opt/ml when running on SageMaker. While we want to
    store checkpoints, we have no interest in uploading them to S3. Therefore, this function changes
    the location to /tmp instead.
    """
    if "SM_MODEL_DIR" in os.environ:  # If running on sagemaker, redirect checkpoints to /tmp
        assert uri.startswith("/opt/ml")
        uri = "/tmp" + uri[7:]
    return uri


def config_space_to_dict(
    config_space: Dict[str, Union[Domain, int, float, str]]
) -> Dict[str, Union[int, float, str]]:
    """Converts `config_space` into a dictionary that can be saved as a json file."""
    # TODO: remove with Syne Tune 0.3.3
    return {k: to_dict(v) if isinstance(v, Domain) else v for k, v in config_space.items()}


def config_space_from_dict(
    config_space_dict: Dict[str, Union[int, float, str]]
) -> Dict[str, Union[Domain, int, float, str]]:
    """Converts the given dictionary into a Syne Tune search space."""
    # TODO: remove with Syne Tune 0.3.3
    return {k: from_dict(v) if isinstance(v, dict) else v for k, v in config_space_dict.items()}


def best_hyperparameters(
    experiment: ExperimentResult, config_space: Dict[str, Union[Domain, int, float, str]]
) -> Dict[str, Union[int, float, str]]:
    """Returns the values of all keys in the `config_space` that belong to a Syne Tune search
    space."""
    return {
        k[7:]: v
        for k, v in experiment.best_config().items()
        if k.startswith("config_") and isinstance(config_space[k[7:]], Domain)
    }


def is_syne_tune_config_space(config_space: Dict[str, Any]) -> bool:
    """Returns `True` if any value in the configuration space defines a Syne Tune search space."""
    return any(
        [
            isinstance(hyperparameter_instance, Domain)
            for hyperparameter_instance in config_space.values()
        ]
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import importlib.util
import sys
import warnings
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torchmetrics
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler

from renate import defaults
from renate.benchmark.scenarios import Scenario
from renate.data.data_module import RenateDataModule
from renate.evaluation.evaluator import evaluate
from renate.models import RenateModule


def evaluate_and_record_results(
    results: Dict[str, List[List[float]]],
    model: RenateModule,
    data_module: Union[Scenario, RenateDataModule],
    transform: Optional[Callable] = None,
    target_transform: Optional[Callable] = None,
    task_id: str = defaults.TASK_ID,
    logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
    metric_postfix: str = "",
    batch_size: int = defaults.BATCH_SIZE,
    accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
    devices: Optional[int] = None,
    strategy: str = defaults.DISTRIBUTED_STRATEGY,
    precision: str = defaults.PRECISION,
) -> Dict[str, List[List[float]]]:
    """A helper function that performs the evaluation on test data and records quantitative metrics
    in a dictionary.

    Args:
        results: The results dictionary to which the results should be saved.
        model: A RenateModule to be evaluated.
        data_module: A Scenario or RenateDataModule from which the test data is queried.
        transform: The transformation applied for evaluation.
        target_transform: The target transformation applied for evaluation.
        task_id: The task ID for which the evaluation should be performed.
        logged_metrics: Metrics logged additional to the default ones.
        metric_postfix: The postfix for the metric names.
        batch_size: A batch size for the test loader.
        accelerator: Accelerator used by PyTorch Lightning to train the model.
        devices: Devices used by PyTorch Lightning to train the model. If the devices flag is not
            defined, it will assume devices to be "auto" and fetch the `auto_device_count` from the
            `accelerator`.
        strategy: Name of the distributed training strategy to use.
            `More details <https://lightning.ai/docs/pytorch/stable/extensions/strategy.html>`__
        precision: Type of bit precision to use.
            `More details <https://lightning.ai/docs/pytorch/stable/common/precision_basic.html>`__
    """
    update_results = evaluate(
        model=model,
        test_dataset=data_module.test_data(),
        test_collate_fn=data_module.test_collate_fn(),
        task_id=task_id,
        batch_size=batch_size,
        transform=transform,
        target_transform=target_transform,
        logged_metrics=logged_metrics,
        accelerator=accelerator,
        devices=devices,
        strategy=strategy,
        precision=precision,
    )
    for key, value in update_results.items():
        result_key = f"{key}{metric_postfix}"
        if result_key not in results:
            results[result_key] = []
        results[result_key].append(value)
    return results


def get_model(config_module: ModuleType, **kwargs: Any) -> RenateModule:
    """Creates and returns a model instance."""
    return getattr(config_module, "model_fn")(**kwargs)


def get_data_module(config_module: ModuleType, **kwargs: Any) -> RenateDataModule:
    """Creates and returns a data module instance."""
    return getattr(config_module, "data_module_fn")(**kwargs)


def _convert_loss(loss_fn: torch.nn.Module):
    """Changes PyTorch loss such that it uses no reduction."""
    if hasattr(loss_fn, "reduction") and loss_fn.reduction != "none":
        warnings.warn(
            "Renate assumes that your loss function returns a loss value for each data point."
            f"Your loss function uses reduction={loss_fn.reduction}, changing to `none`."
        )
        loss_fn.reduction = "none"


def get_loss_fn(config_module: ModuleType, convert: bool, **kwargs: Any) -> torch.nn.Module:
    """Creates and returns the loss function from config"""
    loss_fn = getattr(config_module, "loss_fn")(**kwargs)
    if convert:
        _convert_loss(loss_fn)
    return loss_fn


def get_optimizer(
    config_module: ModuleType, **kwargs: Any
) -> Optional[Callable[[List[Parameter]], Optimizer]]:
    """Creates partial optimizer object from config."""
    optimizer_fn_name = "optimizer_fn"
    if optimizer_fn_name in vars(config_module):
        return getattr(config_module, optimizer_fn_name)(**kwargs)


def get_learning_rate_scheduler(
    config_module: ModuleType, **kwargs: Any
) -> Optional[
    Tuple[Callable[[Optimizer], _LRScheduler], defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE]
]:
    """Creates partial learning rate scheduler object from config."""
    lr_scheduler_fn_name = "lr_scheduler_fn"
    if lr_scheduler_fn_name in vars(config_module):
        return getattr(config_module, lr_scheduler_fn_name)(**kwargs)


def get_metrics(
    config_module: ModuleType, **kwargs: Any
) -> Optional[Dict[str, torchmetrics.Metric]]:
    """Creates and returns a dictionary of metrics."""
    metrics_fn_name = "metrics_fn"
    if metrics_fn_name in vars(config_module):
        return getattr(config_module, metrics_fn_name)(**kwargs)


def get_and_prepare_data_module(config_module: ModuleType, **kwargs: Any) -> RenateDataModule:
    """Prepares data."""
    data_module = get_data_module(config_module, **kwargs)
    data_module.prepare_data()
    return data_module


def get_and_setup_data_module(
    config_module: ModuleType,
    prepare_data: bool,
    **kwargs: Any,
) -> RenateDataModule:
    """Creates data module and possibly calls the prepare_data function needed for setup"""
    data_module = get_data_module(config_module, **kwargs)
    if prepare_data:
        data_module.prepare_data()
    data_module.setup()
    return data_module


def import_module(module_name: str, location: str) -> ModuleType:
    """Imports Python module from file location."""
    spec = importlib.util.spec_from_file_location(name=module_name, location=location)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from abc import ABC
from pathlib import Path
from typing import Optional, Union

from renate import defaults
from renate.data.data_module import RenateDataModule


class DataIncrementalDataModule(RenateDataModule, ABC):
    """Base class for all :py:class:`~renate.data.data_module.RenateDataModule` compatible with
    :py:class:`~renate.benchmark.scenarios.DataIncrementalScenario`.

    Defines the API required by the :py:class:`~renate.benchmark.scenarios.DataIncrementalScenario`.
    All classes extending this class must load the datasets corresponding to the value in
    ``data_id`` whenever ``setup()`` is called.

    Args:
        data_path: the path to the folder containing the dataset files.
        data_id: Time slice to be loaded.
        src_bucket: the name of the s3 bucket. If not provided, downloads the data from original
            source.
        src_object_name: the folder path in the s3 bucket.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_path: Union[Path, str],
        data_id: Union[int, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super().__init__(
            data_path=data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self.data_id = data_id



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path

import pytest
import torch
from avalanche.training.plugins import ReplayPlugin

from conftest import AVALANCHE_LEARNER_KWARGS
from renate import defaults
from renate.updaters.avalanche.learner import AvalancheICaRLLearner, plugin_by_class
from renate.updaters.avalanche.model_updater import ExperienceReplayAvalancheModelUpdater


def get_model_and_dataset(dataset_size: int = 100):  # TODO: remove
    model, dataset, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=dataset_size,
        test_num_samples=5,
    )
    return model, dataset


@pytest.mark.parametrize("provide_folder", [True, False])
def test_avalanche_model_updater(tmpdir, provide_folder):
    model, train_dataset, test_data = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )
    model_updater = pytest.helpers.get_avalanche_updater(
        model, output_state_folder=defaults.output_state_folder(tmpdir) if provide_folder else None
    )
    y_hat_before_train = model(test_data, task_id=defaults.TASK_ID)
    model_updater.update(train_dataset, task_id=defaults.TASK_ID)
    y_hat_after_train = model(test_data, task_id=defaults.TASK_ID)
    assert y_hat_before_train.shape[0] == y_hat_after_train.shape[0]
    assert not torch.allclose(y_hat_before_train, y_hat_after_train)


@pytest.mark.parametrize("learner_class", list(AVALANCHE_LEARNER_KWARGS.keys()))
def test_continuation_of_training_with_avalanche_model_updater(tmpdir, learner_class):
    if learner_class == AvalancheICaRLLearner:
        return
    model, train_dataset, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=1,
        train_num_samples=10,
        test_num_samples=5,
        add_icarl_class_means=learner_class == AvalancheICaRLLearner,
    )
    state_url = defaults.input_state_folder(tmpdir)
    model_updater = pytest.helpers.get_avalanche_updater(
        model,
        learner_class=learner_class,
        learner_kwargs=AVALANCHE_LEARNER_KWARGS[learner_class],
        output_state_folder=state_url,
        max_epochs=2,
    )
    model = model_updater.update(train_dataset, task_id=defaults.TASK_ID)
    model_updater = pytest.helpers.get_avalanche_updater(
        model,
        learner_class=learner_class,
        learner_kwargs=AVALANCHE_LEARNER_KWARGS[learner_class],
        input_state_folder=state_url,
        max_epochs=2,
    )
    model_updater.update(train_dataset, task_id=defaults.TASK_ID)


@pytest.mark.parametrize(
    "batch_size,memory_size,batch_memory_frac",
    [[20, 10, 0.5], [30, 10, 0.34], [20, 100, 0.5], [10, 30, 0.1], [100, 10, 0.03]],
)
def test_experience_replay_buffer_size(tmpdir, batch_size, memory_size, batch_memory_frac):
    expected_memory_batch_size = int(batch_memory_frac * batch_size)
    expected_batch_size = batch_size - expected_memory_batch_size
    dataset_size = 100
    model, dataset = get_model_and_dataset(dataset_size)
    learner_kwargs = {
        "memory_size": memory_size,
        "batch_memory_frac": batch_memory_frac,
        "batch_size": batch_size,
    }
    model_updater = ExperienceReplayAvalancheModelUpdater(
        output_state_folder=str(Path(tmpdir) / "0"),
        model=model,
        loss_fn=pytest.helpers.get_loss_fn("mean"),
        optimizer=pytest.helpers.get_partial_optimizer(),
        **learner_kwargs,
        max_epochs=1,
        accelerator="cpu",
    )
    model_updater.update(train_dataset=dataset)
    replay_plugin = plugin_by_class(ReplayPlugin, model_updater._learner.plugins)
    assert replay_plugin.batch_size == expected_batch_size
    assert replay_plugin.mem_size == memory_size
    assert replay_plugin.batch_size_mem == expected_memory_batch_size
    assert len(replay_plugin.storage_policy.buffer) == min(
        memory_size, dataset_size, len(replay_plugin.storage_policy.buffer)
    )
    assert len(model_updater._learner.dataloader.data) == dataset_size
    for i in range(6):
        del model_updater
        del replay_plugin
        _, dataset = get_model_and_dataset(dataset_size)
        model_updater = ExperienceReplayAvalancheModelUpdater(
            input_state_folder=str(Path(tmpdir) / str(i)),
            output_state_folder=str(Path(tmpdir) / str(i + 1)),
            model=model,
            loss_fn=pytest.helpers.get_loss_fn("mean"),
            optimizer=pytest.helpers.get_partial_optimizer(),
            **learner_kwargs,
            max_epochs=1,
        )
        replay_plugin = plugin_by_class(ReplayPlugin, model_updater._learner.plugins)

        assert replay_plugin.batch_size == expected_batch_size
        assert replay_plugin.mem_size == memory_size
        assert replay_plugin.batch_size_mem == expected_memory_batch_size
        model_updater.update(train_dataset=dataset)
        assert len(model_updater._learner.dataloader.data) == dataset_size
        assert len(model_updater._learner.dataloader.memory) == min(
            memory_size, (i + 1) * dataset_size
        )
    assert len(replay_plugin.storage_policy.buffer) == min(memory_size, 2 * dataset_size)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
from torch.nn import Linear
from torch.optim import AdamW, SGD
from torch.optim.lr_scheduler import CosineAnnealingLR, StepLR
from torchmetrics.classification import MulticlassAccuracy
from torchvision.transforms import Compose, Normalize, ToTensor

from renate.benchmark import experiment_config
from renate.benchmark.datasets.nlp_datasets import HuggingFaceTextDataModule, MultiTextDataModule
from renate.benchmark.datasets.vision_datasets import (
    CDDBDataModule,
    CLEARDataModule,
    TorchVisionDataModule,
)
from renate.benchmark.experiment_config import (
    data_module_fn,
    get_data_module,
    get_scenario,
    loss_fn,
    lr_scheduler_fn,
    metrics_fn,
    model_fn,
    models,
    optimizer_fn,
    train_transform,
)
from renate.benchmark.scenarios import (
    ClassIncrementalScenario,
    DataIncrementalScenario,
    FeatureSortingScenario,
    HueShiftScenario,
    IIDScenario,
    ImageRotationScenario,
    PermutationScenario,
)
from renate.models.prediction_strategies import ICaRLClassificationStrategy


@pytest.mark.slow
@pytest.mark.parametrize(
    "model_name,expected_model_class",
    [(model_name, model_class) for model_name, model_class in models.items()],
)
def test_model_fn(model_name, expected_model_class):
    model = model_fn(
        model_state_url=None,
        model_name=model_name,
        num_inputs=1 if model_name == "MultiLayerPerceptron" else None,
        num_outputs=2,
        num_hidden_layers=1 if model_name == "MultiLayerPerceptron" else None,
        hidden_size=1 if model_name == "MultiLayerPerceptron" else None,
        pretrained_model_name_or_path="distilbert-base-uncased"
        if model_name == "HuggingFaceTransformer"
        else None,
    )
    assert isinstance(model, expected_model_class)


@pytest.mark.parametrize(
    "dataset_name,expected_in_channels",
    (
        ("FashionMNIST", 1),
        ("MNIST", 1),
        ("yearbook", 1),
        ("CIFAR10", 3),
        ("CIFAR100", 3),
        ("fmow", 3),
    ),
)
def test_model_fn_automatic_input_channel_detection_resnet(dataset_name, expected_in_channels):
    """Tests if ResNet architectures input channels are correctly adapted to the dataset."""
    model = model_fn(
        num_outputs=10,
        model_state_url=None,
        model_name="ResNet18",
        dataset_name=dataset_name,
    )
    assert model.get_backbone().conv1.in_channels == expected_in_channels


def test_model_fn_fails_for_unknown_model():
    unknown_model_name = "UNKNOWN_MODEL_NAME"
    with pytest.raises(ValueError, match=f"Unknown model `{unknown_model_name}`"):
        model_fn(num_outputs=10, model_name=unknown_model_name)


@pytest.mark.parametrize(
    "dataset_name,data_module_class,pretrained_model_name_or_path,input_column,target_column",
    (
        ("CIFAR10", TorchVisionDataModule, None, None, None),
        ("CLEAR10", CLEARDataModule, None, None, None),
        (
            "hfd-rotten-tomatoes",
            HuggingFaceTextDataModule,
            "distilbert-base-uncased",
            "text",
            "label",
        ),
        ("MultiText", MultiTextDataModule, "distilbert-base-uncased", None, None),
        ("CDDB", CDDBDataModule, None, None, None),
    ),
)
def test_get_data_module(
    tmpdir,
    dataset_name,
    data_module_class,
    pretrained_model_name_or_path,
    input_column,
    target_column,
):
    data_module = get_data_module(
        data_path=tmpdir,
        dataset_name=dataset_name,
        val_size=0.5,
        seed=0,
        src_bucket=None,
        src_object_name=None,
        pretrained_model_name_or_path=pretrained_model_name_or_path,
        input_column=input_column,
        target_column=target_column,
    )
    assert isinstance(data_module, data_module_class)


def test_get_data_module_fails_for_unknown_dataset(tmpdir):
    unknown_dataset_name = "UNKNOWN_DATASET_NAME"
    with pytest.raises(ValueError, match=f"Unknown dataset `{unknown_dataset_name}`"):
        get_data_module(
            data_path=tmpdir,
            dataset_name=unknown_dataset_name,
            val_size=0.5,
            seed=0,
            src_bucket=None,
            src_object_name=None,
            pretrained_model_name_or_path=None,
            input_column=None,
            target_column=None,
        )


def test_get_scenario_fails_for_unknown_scenario(tmpdir):
    data_module = get_data_module(
        data_path=tmpdir,
        dataset_name="MNIST",
        val_size=0.5,
        seed=0,
        src_bucket=None,
        src_object_name=None,
        pretrained_model_name_or_path=None,
        input_column=None,
        target_column=None,
    )
    unknown_scenario_name = "UNKNOWN_SCENARIO_NAME"
    with pytest.raises(ValueError, match=f"Unknown scenario `{unknown_scenario_name}`"):
        get_scenario(
            scenario_name=unknown_scenario_name, data_module=data_module, chunk_id=0, seed=0
        )


@pytest.mark.parametrize(
    "scenario_name,dataset_name,scenario_kwargs,expected_scenario_class,expected_num_tasks",
    (
        (
            "ClassIncrementalScenario",
            "hfd-trec",
            {
                "pretrained_model_name_or_path": "distilbert-base-uncased",
                "input_column": "text",
                "target_column": "coarse_label",
                "groupings": ((0, 1), (2, 3), (4, 5)),
            },
            ClassIncrementalScenario,
            3,
        ),
        (
            "IIDScenario",
            "MNIST",
            {"num_tasks": 3},
            IIDScenario,
            3,
        ),
        (
            "ImageRotationScenario",
            "MNIST",
            {"degrees": (0, 90, 180)},
            ImageRotationScenario,
            3,
        ),
        (
            "PermutationScenario",
            "MNIST",
            {"num_tasks": 3, "input_dim": (1, 28, 28)},
            PermutationScenario,
            3,
        ),
        (
            "FeatureSortingScenario",
            "MNIST",
            {
                "num_tasks": 5,
                "feature_idx": 0,
                "randomness": 0.3,
            },
            FeatureSortingScenario,
            5,
        ),
        (
            "HueShiftScenario",
            "CIFAR10",
            {"num_tasks": 3, "randomness": 0.5},
            HueShiftScenario,
            3,
        ),
        ("DataIncrementalScenario", "CLEAR10", {"num_tasks": 5}, DataIncrementalScenario, 5),
        (
            "DataIncrementalScenario",
            "arxiv",
            {"num_tasks": 3, "pretrained_model_name_or_path": "distilbert-base-uncased"},
            DataIncrementalScenario,
            3,
        ),
        (
            "DataIncrementalScenario",
            "fmow",
            {},
            DataIncrementalScenario,
            16,
        ),
        (
            "DataIncrementalScenario",
            "DomainNet",
            {"data_ids": ("clipart", "infograph")},
            DataIncrementalScenario,
            2,
        ),
        (
            "DataIncrementalScenario",
            "DomainNet",
            {"groupings": (("clipart", "infograph"), "painting")},
            DataIncrementalScenario,
            2,
        ),
        (
            "DataIncrementalScenario",
            "CDDB",
            {"data_ids": ("biggan", "wild")},
            DataIncrementalScenario,
            2,
        ),
        (
            "DataIncrementalScenario",
            "Core50",
            {"data_ids": (2, 3)},
            DataIncrementalScenario,
            2,
        ),
    ),
    ids=[
        "class_incremental",
        "iid",
        "rotation",
        "permutation",
        "feature_sorting",
        "hue_shift",
        "data_incremental with CLEAR",
        "wild_time_text_with_tokenizer",
        "wild_time_image_all_tasks",
        "domainnet_by data_id",
        "domainnet by groupings",
        "cddb by dataid",
        "core50 by dataid",
    ],
)
@pytest.mark.parametrize("val_size", (0, 0.5), ids=["no_val", "val"])
def test_data_module_fn(
    tmpdir,
    scenario_name,
    dataset_name,
    scenario_kwargs,
    expected_scenario_class,
    expected_num_tasks,
    val_size,
):
    scenario = data_module_fn(
        data_path=tmpdir,
        chunk_id=0,
        seed=0,
        scenario_name=scenario_name,
        dataset_name=dataset_name,
        val_size=val_size,
        **scenario_kwargs,
    )
    assert isinstance(scenario, expected_scenario_class)
    if expected_scenario_class == ClassIncrementalScenario:
        assert scenario._class_groupings == scenario_kwargs["groupings"]
    elif expected_scenario_class == FeatureSortingScenario:
        assert scenario._feature_idx == scenario_kwargs["feature_idx"]
        assert scenario._randomness == scenario_kwargs["randomness"]
    elif expected_scenario_class == HueShiftScenario:
        assert scenario._randomness == scenario_kwargs["randomness"]
    elif expected_scenario_class == DataIncrementalScenario:
        if "pretrained_model_name_or_path" in scenario_kwargs:
            assert scenario._data_module._tokenizer is not None
        elif dataset_name not in ["CLEAR10", "CLEAR100", "DomainNet", "CDDB", "Core50"]:
            assert scenario._data_module._tokenizer is None
    assert scenario._num_tasks == expected_num_tasks


@pytest.mark.parametrize(
    "dataset_name,expected_train_transform_class, expected_test_transform_class,model_name",
    (
        ("MNIST", type(None), type(None), "ResNet18CIFAR"),
        ("FashionMNIST", type(None), type(None), "ResNet18CIFAR"),
        ("CIFAR10", Compose, Normalize, "ResNet18CIFAR"),
        ("CIFAR100", Compose, Normalize, "ResNet18CIFAR"),
        ("CLEAR10", Compose, Compose, "ResNet18"),
        ("CDDB", Compose, Compose, None),
        ("Core50", Compose, Compose, None),
        ("DomainNet", Compose, Compose, "VisionTransformerB16"),
        ("hfd-rotten_tomatoes", type(None), type(None), "HuggingFaceTransformer"),
        ("fmow", Compose, Compose, "ResNet18"),
        ("yearbook", ToTensor, ToTensor, "ResNet18CIFAR"),
        ("yearbook", Compose, Compose, "VisionTransformerB16"),
    ),
)
def test_transforms(
    dataset_name, expected_train_transform_class, expected_test_transform_class, model_name
):
    train_preprocessing = train_transform(dataset_name, model_name)
    test_preprocessing = experiment_config.test_transform(dataset_name, model_name)
    assert isinstance(train_preprocessing, expected_train_transform_class)
    assert isinstance(test_preprocessing, expected_test_transform_class)


def test_transforms_fails_for_unknown_dataset():
    unknown_dataset_set = "UNKNOWN_DATASET_NAME"
    for transform_function in [train_transform, experiment_config.test_transform]:
        with pytest.raises(ValueError, match=f"Unknown dataset `{unknown_dataset_set}`"):
            transform_function(unknown_dataset_set)


@pytest.mark.parametrize(
    "learning_rate_scheduler,expected_lr_class,expected_interval",
    (
        ("StepLR", StepLR, "epoch"),
        ("CosineAnnealingLR", CosineAnnealingLR, "step"),
        (None, None, "epoch"),
    ),
)
def test_lr_scheduler_fn(learning_rate_scheduler, expected_lr_class, expected_interval):
    scheduler, interval = lr_scheduler_fn(
        learning_rate_scheduler=learning_rate_scheduler,
        learning_rate_scheduler_interval=expected_interval,
    )
    assert interval == expected_interval
    if learning_rate_scheduler is None:
        assert scheduler is None
    else:
        scheduler = scheduler(SGD(Linear(1, 1).parameters(), lr=0.1))
        assert isinstance(scheduler, expected_lr_class)


def test_lr_scheduler_fn_fails_for_unknown_scheduler():
    unknown_lr_scheduler = "UNKNOWN_SCHEDULER_NAME"
    with pytest.raises(ValueError, match=f"Unknown scheduler `{unknown_lr_scheduler}`."):
        lr_scheduler_fn(unknown_lr_scheduler)


@pytest.mark.slow
@pytest.mark.parametrize("model_name", [model_name for model_name in models])
@pytest.mark.parametrize("updater", ("ER", "Avalanche-iCaRL"))
def test_prediction_strategy_is_correctly_set(model_name, updater):
    """If iCaRL is used, the model prediction strategy must be changed."""
    model_kwargs = {
        "model_name": model_name,
        "updater": updater,
        "num_outputs": 2,
    }
    if model_name == "MultiLayerPerceptron":
        model_kwargs.update({"num_inputs": 10, "hidden_size": 10, "num_hidden_layers": 2})
    elif model_name == "HuggingFaceTransformer":
        model_kwargs["pretrained_model_name_or_path"] = "distilbert-base-uncased"
    if model_name == "HuggingFaceTransformer" and updater == "Avalanche-iCaRL":
        with pytest.raises(ValueError, match="Transformers do not support iCaRL."):
            model_fn(**model_kwargs)
    else:
        model = model_fn(**model_kwargs)
        if updater == "ER":
            assert not hasattr(model, "_prediction_strategy") or model._prediction_strategy is None
        else:
            assert isinstance(model._prediction_strategy, ICaRLClassificationStrategy)


def test_loss_fn_returns_correct_reduction_type():
    assert loss_fn("ER").reduction == "none"
    assert loss_fn("Avalanche-ER").reduction == "mean"


def test_metrics_fn_contains_accuracy():
    assert isinstance(metrics_fn(num_outputs=2)["accuracy"], MulticlassAccuracy)
    assert isinstance(metrics_fn(num_outputs=10)["accuracy"], MulticlassAccuracy)


def test_optimizer_fn():
    expected_learning_rate = 0.12
    partial_optimizer = optimizer_fn(
        optimizer="AdamW", learning_rate=expected_learning_rate, weight_decay=0.1
    )
    optimizer: AdamW = partial_optimizer(Linear(10, 10).parameters())
    assert isinstance(optimizer, AdamW)
    assert optimizer.defaults["lr"] == expected_learning_rate


def test_optimizer_fn_unknown_optimizer():
    assert optimizer_fn(optimizer="UNKNOWN_OPTIMIZER", learning_rate=0.1, weight_decay=1) is None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import pytest

from renate.models.layers.shared_linear import SharedMultipleLinear


@pytest.mark.parametrize("share_parameters", [True, False])
def test_shared_multiple_classifier(share_parameters):
    model = SharedMultipleLinear(5, 3, share_parameters=share_parameters, num_updates=10)
    num_elems = sum(x.numel() for x in model.parameters())
    assert num_elems == [10 * 5 * 3 + 10 * 3, 5 * 3 + 3][share_parameters]

    model.increment_task()
    num_elems = sum(x.numel() for x in model.parameters())
    assert num_elems == [(10 + 1) * 5 * 3 + (10 + 1) * 3, 5 * 3 + 3][share_parameters]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple

import torch
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset

from renate import defaults
from renate.memory import DataBuffer
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.learner import Learner, ReplayLearner
from renate.updaters.model_updater import ModelUpdater
from renate.utils.pytorch import move_tensors_to_device, reinitialize_model_parameters


def double_distillation_loss(
    predicted_logits: torch.Tensor, target_logits: torch.Tensor
) -> torch.Tensor:
    """Double distillation loss, where target logits are normalized across the class-dimension.

    This normalization is useful when distilling from multiple teachers and was proposed in

        TODO: Fix citation once we agreed on a format.
        Zhang, Junting, et al. "Class-incremental learning via deep model consolidation."
        Proceedings of the IEEE/CVF Winter Conference on Applications of Computer Vision. 2020.

    Args:
        predicted_logits: Logit predictions of the student model, size `(B, C)`, where `B` is the
            batch size and `C` is the number of classes.
        target_logits: Logits obtained from the teacher model(s), same size `(B, C)`.

    Returns:
        A tensor of size `(B,)` containing the loss values for each datapoint in the batch.
    """
    target_logits_normalized = target_logits - target_logits.mean(dim=1, keepdim=True)
    return 0.5 * (predicted_logits - target_logits_normalized).pow(2).mean(dim=1)


@torch.no_grad()
def extract_logits(
    model: torch.nn.Module,
    dataset: Dataset,
    batch_size: int,
    task_id: Optional[str] = defaults.TASK_ID,
) -> torch.Tensor:
    """Extracts logits from a model for each point in a dataset.

    Args:
        model: The model. `model.get_logits(X)` is assumed to return logits.
        dataset: The dataset.
        batch_size: Batch size used to iterate over the dataset.
        task_id: Task id to be used, e.g., to select the output head.

    Returns:
        A tensor `logits` of shape `(N, C)` where `N` is the length of the dataset and `C` is
        the output dimension of `model`, i.e., the number of classes.
    """
    logits = []
    loader = DataLoader(dataset, batch_size, shuffle=False, drop_last=False, pin_memory=True)
    for batch in loader:
        inputs = batch[0][0] if isinstance(dataset, DataBuffer) else batch[0]
        device = next(model.parameters()).device
        inputs = move_tensors_to_device(inputs, device)
        logits.append(model.get_logits(inputs, task_id))
    return torch.cat(logits, dim=0)


class RepeatedDistillationModelUpdater(ModelUpdater):
    """Repeated Distillation (RD) is inspired by Deep Model Consolidation (DMC), which was proposed
    in

        TODO: Fix citation once we agreed on a format.
        Zhang, Junting, et al. "Class-incremental learning via deep model consolidation."
        Proceedings of the IEEE/CVF Winter Conference on Applications of Computer Vision. 2020.

    The idea underlying RD is the following: Given a new task/batch, a new model copy is trained
    from scratch on that data. Subsequently, this expert model is consolidated with the previous
    model state via knowledge distillation. The resulting consolidated model state is maintained,
    whereas the expert model is discarded.

    Our variant differs from the original algorithm in two ways:
        - The original algorithm is designed specifically for the class-incremental setting, where
          each new task introduces one or more novel classes. This variant is designed for the
          general continual learning setting with a pre-determined number of classes.
        - The original method is supposed to be memory-free and uses auxiliary data for the model
          consolidation phase. Our variant performs knowledge distillation over a memory
    """

    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        memory_size: int,
        learning_rate_scheduler: Optional[partial] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        batch_size: int = defaults.BATCH_SIZE,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: str = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        seed: Optional[int] = None,
        early_stopping_enabled=False,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        learner_kwargs = {"memory_size": memory_size, "batch_size": batch_size, "seed": seed}
        super().__init__(
            model=model,
            loss_fn=loss_fn,
            optimizer=optimizer,
            learner_class=RepeatedDistillationLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            learning_rate_scheduler=learning_rate_scheduler,
            learning_rate_scheduler_interval=learning_rate_scheduler_interval,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            strategy=strategy,
            precision=precision,
            early_stopping_enabled=early_stopping_enabled,
            logged_metrics=logged_metrics,
            deterministic_trainer=deterministic_trainer,
            gradient_clip_algorithm=gradient_clip_algorithm,
            gradient_clip_val=gradient_clip_val,
            mask_unused_classes=mask_unused_classes,
        )

    def update(
        self,
        train_dataset: Dataset,
        val_dataset: Optional[Dataset] = None,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> RenateModule:
        """Updates the model using the data passed as input.

        Args:
            train_dataset: The training data.
            val_dataset: The validation data.
            train_dataset_collate_fn: collate_fn used to merge a list of samples to form a
                mini-batch of Tensors for the training data.
            val_dataset_collate_fn: collate_fn used to merge a list of samples to form a
                mini-batch of Tensors for the validation data.
            task_id: The task id.
        """
        # First, train a copy of the model on the new data from scratch as an expert model. We use
        # the base `Learner` for that. The expert model and learner do not need to persist, we only
        # need it to extract logits.
        expert_model = copy.deepcopy(self._model)
        reinitialize_model_parameters(expert_model)
        expert_learner = Learner(
            model=expert_model,
            train_transform=self._train_transform,
            train_target_transform=self._train_target_transform,
            **{key: value for key, value in self._learner_kwargs.items() if key != "memory_size"},
        )
        expert_learner.on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._fit_learner(expert_learner)

        # Extract logits from the expert model and register them with the consolidation learner.
        expert_logits = extract_logits(
            expert_model,
            train_dataset,
            batch_size=self._learner_kwargs["batch_size"],
            task_id=task_id,
        )

        self._learner.update_expert_logits(expert_logits)
        del expert_model
        del expert_learner

        # Run consolidation.
        self._learner.on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._fit_learner(self._learner)
        return self._model


class RepeatedDistillationLearner(ReplayLearner):
    """A learner performing distillation."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._expert_logits: Optional[torch.Tensor] = None

    def update_expert_logits(self, new_expert_logits: torch.Tensor) -> None:
        """Update expert logits."""
        self._expert_logits = new_expert_logits

    def on_model_update_start(
        self,
        train_dataset: Dataset,
        val_dataset: Dataset,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[int] = None,
    ) -> None:
        """Called before a model update starts."""
        super().on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._memory_buffer.update(train_dataset, metadata={"logits": self._expert_logits})
        reinitialize_model_parameters(self._model)
        self._expert_logits = None

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self._memory_buffer,
            batch_size=self._batch_size,
            shuffle=True,
            generator=self._rng,
            pin_memory=True,
            collate_fn=self._train_collate_fn,
        )

    def on_model_update_end(self) -> None:
        """Called right before a model update terminates."""
        # Update the logits in memory using the newly consolidated model.
        logits = extract_logits(self._model, self._memory_buffer, self._batch_size)
        self._memory_buffer.set_metadata("logits", logits)

    def training_step(
        self,
        batch: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        batch_idx: int,
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        (inputs, targets), metadata = batch
        outputs = self(inputs)
        loss = double_distillation_loss(outputs, metadata["logits"]).mean()
        self._update_metrics(outputs, targets, prefix="train")
        self._loss_collections["train_losses"]["base_loss"](loss)
        return {"loss": loss, "outputs": outputs}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest

from conftest import SAMPLE_CLASSIFICATION_RESULTS
from renate.evaluation.metrics.classification import (
    average_accuracy,
    backward_transfer,
    forgetting,
    forward_transfer,
    micro_average_accuracy,
)


@pytest.mark.parametrize(
    "task_id,result",
    [
        [0, 0.9362000226974487],
        [1, sum([0.8284000158309937, 0.9506999850273132]) / 2],
        [2, sum([0.4377000033855438, 0.48260000348091125, 0.9438999891281128]) / 3],
    ],
)
def test_average_accuracy(task_id, result):
    results = SAMPLE_CLASSIFICATION_RESULTS
    assert pytest.approx(average_accuracy(results, task_id, [])) == result


@pytest.mark.parametrize(
    "task_id,expected_result,num_instances",
    [
        [0, 0.9362000226974487, [10, 5, 10]],
        [1, 0.8691666722297668, [10, 5, 10]],
        [2, 0.6491599977016449, [10, 5, 10]],
    ],
)
def test_micro_average_accuracy(task_id, expected_result, num_instances):
    results = SAMPLE_CLASSIFICATION_RESULTS
    assert pytest.approx(micro_average_accuracy(results, task_id, num_instances)) == expected_result


@pytest.mark.parametrize(
    "task_id,result",
    [
        [0, 0.0],
        [1, (1 / 1) * (0.9362000226974487 - 0.8284000158937)],
        [
            2,
            (1 / 2)
            * sum(
                [0.9362000226974487 - 0.4377000033855438, 0.9506999850273132 - 0.48260000348091125]
            ),
        ],
    ],
)
def test_forgetting(task_id, result):
    results = SAMPLE_CLASSIFICATION_RESULTS
    assert pytest.approx(forgetting(results, task_id, [])) == result


@pytest.mark.parametrize(
    "task_id,result",
    [
        [0, 0.0],
        [1, (1 / 1) * (0.8284000158309937 - 0.9362000226974487)],
        [
            2,
            (1 / 2)
            * sum(
                [0.4377000033855438 - 0.9362000226974487, 0.48260000348091125 - 0.9506999850273132]
            ),
        ],
    ],
)
def test_backward_transfer(task_id, result):
    results = SAMPLE_CLASSIFICATION_RESULTS
    assert pytest.approx(backward_transfer(results, task_id, [])) == result


@pytest.mark.parametrize(
    "task_id,result",
    [
        [0, 0.0],
        [1, (1 / 1) * (0.6093000173568726 - 0.1)],
        [
            2,
            (1 / 2) * sum([0.6093000173568726 - 0.1, 0.3382999897003174 - 0.09]),
        ],
    ],
)
def test_forward_transfer(task_id, result):
    results = SAMPLE_CLASSIFICATION_RESULTS
    assert pytest.approx(forward_transfer(results, task_id, [])) == result



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch
from torch.utils.data import Dataset

from renate.memory.storage import FileTensorStorage


def nested_tensors_equal(t1, t2):
    if type(t1) is not type(t2):
        return False
    if isinstance(t1, torch.Tensor):
        return torch.equal(t1, t2)
    if isinstance(t1, (tuple, list)):
        if len(t1) != len(t2):
            return False
        return all(nested_tensors_equal(t1_, t2_) for t1_, t2_ in zip(t1, t2))
    if isinstance(t1, dict):
        if set(t1.keys()) != set(t2.keys()):
            return False
        return all(nested_tensors_equal(t1[key], t2[key]) for key in t1.keys())


def make_dataset_same_sizes(dataset_length, return_type):
    class CustomDataset(Dataset):
        def __init__(self, x, y, return_type="tensor") -> None:
            self.x, self.y = x, y
            self.return_type = return_type

        def __getitem__(self, index):
            if self.return_type == "tensor":
                return self.x[index]
            elif self.return_type == "tuple":
                return (self.x[index], self.y[index])
            elif self.return_type == "dict":
                return {"x": self.x[index], "y": self.y[index]}
            elif self.return_type == "list":
                return [self.x[index], self.y[index]]

        def __len__(self):
            return self.x.shape[0]

    data_x = torch.rand(dataset_length, 3, 32, 32)
    data_y = torch.randint_like(data_x, 5)

    return CustomDataset(data_x, data_y, return_type=return_type)


def make_dataset_different_sizes(dataset_length, return_type):
    class CustomDataset(Dataset):
        def __init__(self, return_type="tensor") -> None:
            self.return_type = return_type

        def __getitem__(self, index):
            x = torch.ones(index + 2, index + 3).float() * (index + 1)
            y = torch.ones(index + 2, index + 3).int() * (index + 2)
            if self.return_type == "tensor":
                return x
            elif self.return_type == "tuple":
                return (x, y)
            elif self.return_type == "dict":
                return {"x": x, "y": y}
            elif self.return_type == "list":
                return [x, y]

        def __len__(self):
            return dataset_length

    return CustomDataset(return_type=return_type)


@pytest.mark.parametrize("length", [1, 10])
@pytest.mark.parametrize("return_type", ["tuple", "dict", "tensor", "list"])
@pytest.mark.parametrize(
    "dataset_maker_fn", [make_dataset_same_sizes, make_dataset_different_sizes]
)
def test_memory_storage_different_sizes(tmpdir, length, return_type, dataset_maker_fn):
    ds = dataset_maker_fn(length, return_type)
    storage = FileTensorStorage(tmpdir)
    storage.dump_dataset(ds)

    del storage
    storage = FileTensorStorage(tmpdir)
    for i in range(length):
        assert nested_tensors_equal(storage[i], ds[i])



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict, List


def average_accuracy(
    results: Dict[str, List[List[float]]], task_id: int, num_instances: List[int]
) -> float:
    """Compute the average accuracy of a model.

    This measure is defined by:

    .. math::
        \\frac{1}{T}  sum_{i=1}^T  a_{T,i}

    where :math:`T` is the number of tasks, :math:`a_{T,i}` is the accuracy
    of the model on task :math:`i`, while having learned all tasks up to :math:`T`.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
        num_instances: Count of test data points for each update.
    """
    return sum(results["accuracy"][task_id][: task_id + 1]) / (task_id + 1)


def micro_average_accuracy(
    results: Dict[str, List[List[float]]], task_id: int, num_instances: List[int]
) -> float:
    """Compute the micro average accuracy of a model.

    This measure is defined by the number of correctly classified data points divided by the
    total number of data points. If the number of data points is the same in each update step,
    this is the same as ``average_accuracy``.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
        num_instances: Count of test data points for each update.
    """
    total_num_instances = sum(num_instances[: task_id + 1])
    return sum(
        [
            num_instances[i] / total_num_instances * results["accuracy"][task_id][i]
            for i in range(task_id + 1)
        ]
    )


def forgetting(
    results: Dict[str, List[List[float]]], task_id: int, num_instances: List[int]
) -> float:
    """Compute the forgetting measure of the model.

    This measure is defined by:

    .. math::
        \\frac{1}{T-1}  sum_{i=1}^{T-1}  f_{T,i}

    where :math:`f_{j,i}` is defined as:

    .. math::
        f_{j,i} = \\max_{k\\in\\{1, \\ldots, j=i\\}} a_{k,i} - a_{j,i}

    where :math:`T` is the final task index, :math:`a_{n,i}` is the test classification accuracy
    on task :math:`i` after sequentially learning the nth task and :math:`f_{j,i}` is a measure of
    forgetting on task :math:`i` after training up to task :math:`j`.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
        num_instances: Count of test data points for each update.
    """
    if task_id == 0:
        return 0.0

    def f(results: List[List[float]], j: int, i: int) -> float:
        """A Helper function to compute the: math:`f_{j,i}`."""
        accuracy_ji = results[j][i]
        max_accuracy_ki = max([results[k][i] for k in range(j)])
        return max_accuracy_ki - accuracy_ji

    sum_f = 0.0
    for i in range(task_id):
        sum_f += f(results["accuracy"], task_id, i)
    return sum_f / task_id


def backward_transfer(
    results: Dict[str, List[List[float]]], task_id: int, num_instances: List[int]
) -> float:
    """Compute the backward transfer measure of the model.

    This measure is defined by:

    .. math::
        \\frac{1}{T-1}  sum_{i=1}^{T-1}  a_{T,i} - a_{i,i}

    where :math:`T` is the final task index, :math:`a_{n,i}` is the test classification accuracy
    on task :math:`i` after sequentially learning the nth task.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
        num_instances: Count of test data points for each update.
    """
    if task_id == 0:
        return 0.0
    return (
        sum([results["accuracy"][task_id][i] - results["accuracy"][i][i] for i in range(task_id)])
        / task_id
    )


def forward_transfer(
    results: Dict[str, List[List[float]]], task_id: int, num_instances: List[int]
) -> float:
    """Compute the forward transfer measure of the model.

    This measure is defined by:

    .. math::
        \\frac{1}{T-1}  sum_{i=2}^{T}  a_{i-1,i} - b_{i}

    where :math:`T` is the final task index, :math:`a_{n,i}` is the test classification accuracy
    on task :math:`i` after sequentially learning the nth task. :math:`b_{i}` is the accuracy for
    all :math:`T` tasks recorded at initialisation prior to observing any task.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
        num_instances: Count of test data points for each update.
    """
    if task_id == 0:
        return 0.0
    return (
        sum(
            [
                results["accuracy"][i - 1][i] - results["accuracy_init"][0][i]
                for i in range(1, task_id + 1)
            ]
        )
        / task_id
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
"""
This script is used to launch model run on SageMaker. Previously passed arguments on a different
machine are loaded again and the update process is started with these parameters. Arguments expected
are described in renate.training.run_training_job.
"""
import json

import renate.defaults as defaults
from renate.benchmark.experimentation import execute_experiment_job
from renate.training import run_training_job
from renate.utils.syne_tune import config_space_from_dict

if __name__ == "__main__":
    with open(defaults.JOB_KWARGS_FILE, "r") as f:
        job_kwargs = json.load(f)
    job_kwargs["config_space"] = config_space_from_dict(job_kwargs["config_space"])
    if "experiment_outputs_url" in job_kwargs:
        execute_experiment_job(backend="local", **job_kwargs)
    else:
        run_training_job(backend="local", **job_kwargs)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
"""Sphinx configuration."""

import datetime
import os
import shutil
import sys

sys.path.insert(0, os.path.abspath("../src/"))

import renate  # noqa: E402


def run_apidoc(app):
    """Generate doc stubs using sphinx-apidoc."""
    module_dir = os.path.join(app.srcdir, "../src/")
    output_dir = os.path.join(app.srcdir, "_apidoc")
    excludes = []

    # Ensure that any stale apidoc files are cleaned up first.
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)

    cmd = [
        "--separate",
        "--module-first",
        "--doc-project=API Reference",
        "-o",
        output_dir,
        module_dir,
    ]
    cmd.extend(excludes)

    try:
        from sphinx.ext import apidoc  # Sphinx >= 1.7

        apidoc.main(cmd)
    except ImportError:
        from sphinx import apidoc  # Sphinx < 1.7

        cmd.insert(0, apidoc.__file__)
        apidoc.main(cmd)


def setup(app):
    """Register our sphinx-apidoc hook."""
    app.connect("builder-inited", run_apidoc)


# Sphinx configuration below.
project = "Renate"
version = renate.__version__
release = renate.__version__
copyright = f"2022-{datetime.datetime.now().year}, Amazon"


extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",  # typing shown in api reference
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "hoverxref.extension",  # show reference preview
    "sphinx_copybutton",
    "sphinxext.opengraph",
    "sphinx_paramlinks",  # lightning uses it, we get warnings otherwise
]
coverage_show_missing_items = True

# Make sure the target is unique
autosectionlabel_prefix_document = True

hoverxref_auto_ref = True  # Enable preview for all refs
hoverxref_role_types = {"ref": "tooltip"}

source_suffix = [".rst", ".md"]

master_doc = "index"

autoclass_content = "class"
autodoc_member_order = "bysource"
default_role = "py:obj"

html_theme = "pydata_sphinx_theme"
html_sidebars = {"**": ["sidebar-nav-bs"]}
html_theme_options = {
    "primary_sidebar_end": [],
    "footer_start": ["copyright"],
    "footer_end": [],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/awslabs/Renate",  # required
            "icon": "fa-brands fa-square-github",
            "type": "fontawesome",
        }
    ],
    "use_edit_page_button": True,
    "collapse_navigation": True,
}
html_context = {
    "github_user": "awslabs",
    "github_repo": "Renate",
    "github_version": "dev",
    "doc_path": "doc",
    "default_mode": "light",
}

htmlhelp_basename = "{}doc".format(project)

napoleon_use_rtype = False

rst_prolog = """
.. role:: python(code)
    :language: python
    :class: highlight
"""



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy
import os

import pytest
import torch

from renate.benchmark.models.mlp import MultiLayerPerceptron
from renate.benchmark.models.resnet import ResNet
from renate.benchmark.models.vision_transformer import VisionTransformer
from renate.defaults import TASK_ID
from renate.models import RenateModule
from renate.models.renate_module import RenateWrapper
from renate.utils.deepspeed import recover_object_from_tensor


def test_failing_to_init_abs_class():
    with pytest.raises(TypeError):
        RenateModule({"toy_hyperparam": 1.0}, pytest.helpers.get_loss_fn())


@pytest.mark.parametrize(
    "model",
    [
        pytest.helpers.get_renate_module_mlp(
            num_inputs=10, num_outputs=10, hidden_size=32, num_hidden_layers=3
        ),
        pytest.helpers.get_renate_vision_module(model="resnet", sub_class="resnet18cifar"),
        pytest.helpers.get_renate_vision_module(
            model="visiontransformer", sub_class="visiontransformercifar"
        ),
    ],
)
def test_renate_model_save(tmpdir, model):
    torch.save(model.state_dict(), os.path.join(tmpdir, "test_model.pt"))
    state = torch.load(os.path.join(tmpdir, "test_model.pt"))
    os.remove(os.path.join(tmpdir, "test_model.pt"))

    state["_extra_state"] = recover_object_from_tensor(state["_extra_state"])
    assert "constructor_arguments" in state["_extra_state"].keys()
    assert "tasks_params_ids" in state["_extra_state"].keys()


@pytest.mark.parametrize(
    "test_case,test_cls",
    [
        [
            pytest.helpers.get_renate_module_mlp_and_data(
                num_inputs=10,
                num_outputs=10,
                hidden_size=32,
                num_hidden_layers=3,
                train_num_samples=5,
                test_num_samples=5,
            ),
            MultiLayerPerceptron,
        ],
        [
            pytest.helpers.get_renate_vision_module_and_data(
                model="resnet",
                sub_class="resnet18cifar",
                input_size=(3, 32, 32),
                num_outputs=10,
                train_num_samples=5,
                test_num_samples=5,
            ),
            ResNet,
        ],
        [
            pytest.helpers.get_renate_vision_module_and_data(
                model="visiontransformer",
                sub_class="visiontransformercifar",
                input_size=(3, 32, 32),
                num_outputs=10,
                train_num_samples=5,
                test_num_samples=5,
            ),
            VisionTransformer,
        ],
    ],
)
def test_renate_model_singlehead_save_and_load(tmpdir, test_case, test_cls):
    model, _, test_data = test_case
    model.eval()

    y = torch.randint(3, 8, (5,))
    y_hat_pre_save = model(test_data)
    loss_pre_save = pytest.helpers.get_loss_fn("mean")(y_hat_pre_save, y)

    torch.save(model.state_dict(), os.path.join(tmpdir, "test_model.pt"))
    state = torch.load(os.path.join(tmpdir, "test_model.pt"))
    os.remove(os.path.join(tmpdir, "test_model.pt"))

    model2 = test_cls.from_state_dict(state)
    model2.eval()
    y_hat_post_load = model2(test_data)
    loss_post_load = pytest.helpers.get_loss_fn("mean")(y_hat_post_load, y)

    assert torch.allclose(y_hat_pre_save, y_hat_post_load)
    assert torch.allclose(loss_post_load, loss_pre_save)


@pytest.mark.parametrize(
    "test_case,test_cls",
    [
        [
            pytest.helpers.get_renate_module_mlp_and_data(
                num_inputs=10,
                num_outputs=10,
                hidden_size=32,
                num_hidden_layers=3,
                train_num_samples=5,
                test_num_samples=5,
            ),
            MultiLayerPerceptron,
        ],
        [
            pytest.helpers.get_renate_vision_module_and_data(
                model="resnet",
                sub_class="resnet18cifar",
                input_size=(3, 32, 32),
                num_outputs=10,
                train_num_samples=5,
                test_num_samples=5,
            ),
            ResNet,
        ],
        [
            pytest.helpers.get_renate_vision_module_and_data(
                model="visiontransformer",
                sub_class="visiontransformercifar",
                input_size=(3, 32, 32),
                num_outputs=10,
                train_num_samples=5,
                test_num_samples=5,
            ),
            VisionTransformer,
        ],
    ],
)
def test_renate_model_multihead_save_and_load(tmpdir, test_case, test_cls):
    model, _, test_data = test_case

    model.add_task_params(task_id="first_task")
    model.add_task_params(task_id="second_task")
    model.add_task_params(task_id="third_task")
    model.eval()

    y_hat_first = model(test_data, task_id="first_task")
    y_hat_second = model(test_data, task_id="second_task")
    y_hat_third = model(test_data, task_id="third_task")

    torch.save(model.state_dict(), os.path.join(tmpdir, "test_multihead_model.pt"))
    state = torch.load(os.path.join(tmpdir, "test_multihead_model.pt"))
    os.remove(os.path.join(tmpdir, "test_multihead_model.pt"))

    model2 = test_cls.from_state_dict(state)
    model2.eval()

    y_hat_first_loaded = model2(test_data, task_id="first_task")
    y_hat_second_loaded = model2(test_data, task_id="second_task")
    y_hat_third_loaded = model2(test_data, task_id="third_task")

    assert torch.allclose(y_hat_first_loaded, y_hat_first)
    assert torch.allclose(y_hat_second_loaded, y_hat_second)
    assert torch.allclose(y_hat_third_loaded, y_hat_third)


@pytest.mark.parametrize(
    "test_case",
    [
        pytest.helpers.get_renate_module_mlp_and_data(
            num_inputs=10,
            num_outputs=10,
            hidden_size=32,
            num_hidden_layers=3,
            train_num_samples=5,
            test_num_samples=5,
        ),
        pytest.helpers.get_renate_vision_module_and_data(
            model="resnet",
            sub_class="resnet18cifar",
            input_size=(3, 32, 32),
            num_outputs=10,
            train_num_samples=5,
            test_num_samples=5,
        ),
        pytest.helpers.get_renate_vision_module_and_data(
            model="visiontransformer",
            sub_class="visiontransformercifar",
            input_size=(3, 32, 32),
            num_outputs=10,
            train_num_samples=32,
            test_num_samples=32,
        ),
    ],
)
def test_renate_model_add_params(test_case):
    model, train_dataset, _ = test_case
    model_updater = pytest.helpers.get_simple_updater(model)
    model = model_updater.update(train_dataset=train_dataset, task_id="second_task")
    assert model._tasks_params_ids == {TASK_ID, "second_task"}
    model = model_updater.update(train_dataset=train_dataset, task_id="second_task")
    assert model._tasks_params_ids == {TASK_ID, "second_task"}
    model = model_updater.update(train_dataset=train_dataset, task_id="third_task")
    assert model._tasks_params_ids == {TASK_ID, "second_task", "third_task"}


@pytest.mark.parametrize(
    "test_case",
    [
        pytest.helpers.get_renate_module_mlp_and_data(
            num_inputs=10,
            num_outputs=10,
            hidden_size=32,
            num_hidden_layers=3,
            train_num_samples=5,
            test_num_samples=5,
        ),
        pytest.helpers.get_renate_vision_module_and_data(
            model="resnet",
            sub_class="resnet18cifar",
            input_size=(3, 32, 32),
            num_outputs=10,
            train_num_samples=5,
            test_num_samples=5,
        ),
        pytest.helpers.get_renate_vision_module_and_data(
            model="visiontransformer",
            sub_class="visiontransformercifar",
            input_size=(3, 32, 32),
            num_outputs=10,
            train_num_samples=5,
            test_num_samples=5,
        ),
    ],
)
def test_renate_model_singlehead_single_param_update(test_case):
    model, train_dataset, _ = test_case
    original_model = copy.deepcopy(model)

    model_updater = pytest.helpers.get_simple_updater(model)
    model = model_updater.update(train_dataset=train_dataset, task_id=TASK_ID)
    state_dict = model.state_dict()

    for name, p in original_model.named_parameters():
        assert not torch.equal(p, state_dict[name])


@pytest.mark.parametrize(
    "test_case",
    [
        pytest.helpers.get_renate_module_mlp_and_data(
            num_inputs=10,
            num_outputs=10,
            hidden_size=32,
            num_hidden_layers=3,
            train_num_samples=5,
            test_num_samples=5,
        ),
        pytest.helpers.get_renate_vision_module_and_data(
            model="resnet",
            sub_class="resnet18cifar",
            input_size=(3, 32, 32),
            num_outputs=10,
            train_num_samples=5,
            test_num_samples=5,
        ),
        pytest.helpers.get_renate_vision_module_and_data(
            model="visiontransformer",
            sub_class="visiontransformercifar",
            input_size=(3, 32, 32),
            num_outputs=10,
            train_num_samples=5,
            test_num_samples=5,
        ),
    ],
)
def test_renate_multihead_multi_param_update(test_case):
    model, train_dataset, _ = test_case
    original_model = copy.deepcopy(model)
    model_updater = pytest.helpers.get_simple_updater(model)
    model = model_updater.update(train_dataset=train_dataset, task_id=TASK_ID)
    first_update_model = copy.deepcopy(model)
    model = model_updater.update(train_dataset=train_dataset, task_id="second_task")
    second_update_model = model

    first_update_state_dict = first_update_model.state_dict()
    second_update_state_dict = second_update_model.state_dict()

    for name, p in original_model.named_parameters():
        assert not torch.equal(p, first_update_state_dict[name])
        assert not torch.equal(p, second_update_state_dict[name])

    assert torch.equal(
        first_update_model._tasks_params[TASK_ID].weight,
        second_update_model._tasks_params[TASK_ID].weight,
    )
    assert torch.equal(
        first_update_model._tasks_params[TASK_ID].bias,
        second_update_model._tasks_params[TASK_ID].bias,
    )

    for name, p in first_update_model.named_parameters():
        if TASK_ID in name:
            continue
        assert not torch.equal(p, second_update_state_dict[name])


@pytest.mark.parametrize(
    "torch_model",
    [torch.nn.Sequential(torch.nn.Linear(3, 5), torch.nn.ReLU(), torch.nn.Linear(5, 3))],
)
@torch.no_grad()
def test_renate_wrapper_save_and_load(tmpdir, torch_model):
    renate_module = RenateWrapper(torch_model)
    X = torch.randn(10, 3)
    output_before = renate_module(X)
    assert torch.equal(output_before, torch_model(X))

    torch.save(renate_module.state_dict(), os.path.join(tmpdir, "test_model.pt"))
    del renate_module
    state = torch.load(os.path.join(tmpdir, "test_model.pt"))
    os.remove(os.path.join(tmpdir, "test_model.pt"))

    renate_module = RenateWrapper(torch_model)
    renate_module.load_state_dict(state)

    output_after = renate_module(X)
    assert torch.equal(output_after, output_before)


def test_renate_wrapper_forbids_from_state_dict():
    renate_module = RenateWrapper(torch.nn.Linear(1, 1))
    state_dict = renate_module.state_dict()
    del renate_module
    with pytest.raises(NotImplementedError):
        RenateWrapper.from_state_dict(state_dict)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging

_root_logger = logging.getLogger()
_renate_logger = logging.getLogger(__name__)
_renate_logger.setLevel(logging.INFO)

if not _root_logger.hasHandlers():
    _handler = logging.StreamHandler()
    _handler.setFormatter(
        logging.Formatter("[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s")
    )
    _renate_logger.addHandler(_handler)
    _renate_logger.propagate = False

__version__ = "0.5.1"
